import hashlib
import json
import operator
import uuid
from typing import Any, Dict, Union, overload

import pytz

from commcare_export.jsonpath_utils import split_leftmost
from commcare_export.misc import unwrap, unwrap_val
from commcare_export.repeatable_iterator import RepeatableIterator
from jsonpath_ng import jsonpath
from jsonpath_ng.parser import parse as parse_jsonpath

JSONPATH_CACHE = {}


class CannotBind(Exception):
    pass


class CannotReplace(Exception):
    pass


class CannotEmit(Exception):
    pass


class NotFound(Exception):
    pass


class Env(object):
    """
    An abstract model of an "environment" where data can be bound to
    names and later looked up. Not simply a dictionary as lookup in our
    case may support JsonPath, or may be a chaining of other
    environments, so the abstract interface will allow experimentation
    and customization.
    """

    #
    # Interface
    #
    def bind(self, name: str, value: Any) -> 'Env':
        """
        Returns a new environment that is equivalent to the current
        except the provided key is bound to the value passed in. If the
        environment does not support such a binding, raises  CannotBind
        """
        raise NotImplementedError()

    def lookup(self, key: str) -> Any:
        """
        Note that the return value may be ``None`` which may mean the
        value was unbound or may mean it was found and was None. This
        may need revisiting. This may also raise NotFound if it is the
        sort of environment that does that.
        """
        raise NotImplementedError()

    def replace(self, data: Dict[str, Any]) -> 'Env':
        """
        Completely replace the environment with new data (somewhat like
        "this"-based Map functions a la jQuery). Could be the same as
        creating a new empty env and binding "@" in JsonPath.

        May raise CannotReplace if this environment does not support the
        input replacement
        """
        raise NotImplementedError()

    # Minor impurity of the idea of a binding env: also allow `Emit` to
    # directly call into the environment. It is up to the env whether to
    # store it, write it immediately, or do something clever with
    # iterators, etc.
    def emit_table(self, table_spec):
        raise CannotEmit()

    def has_emitted_tables(self):
        return False

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    #
    # Fluent interface to combinators
    #
    def __or__(self, other):
        return OrElse(self, other)


#
# Combinators
#


class OrElse(Env):
    """
    An environment that chains together a left environment and a right
    environment. Note that this differs from just a bunch of bindings,
    as the two envs might have entirely different mechanisms (for
    example a magic environment for special operators vs a JsonPathEnv
    that always returns a list and operates only on simple data)
    """

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def bind(self, name, value):
        try:
            return OrElse(self.left.bind(name, value), self.right)
        except CannotBind:
            return OrElse(self.left, self.right.bind(name, value))

    def lookup(self, name):
        try:
            return self.left.lookup(name)
        except NotFound:
            return self.right.lookup(name)

    def replace(self, data):
        # A bit sketchy...
        try:
            return OrElse(self.left.replace(data), self.right)
        except CannotReplace:
            return OrElse(self.left, self.right.replace(data))

    def emit_table(self, table_spec):
        try:
            return self.left.emit_table(table_spec)
        except CannotEmit:
            return self.right.emit_table(table_spec)

    def has_emitted_tables(self):
        return any([
            self.left.has_emitted_tables(),
            self.right.has_emitted_tables()
        ])

    def __enter__(self):
        self.left.__enter__()
        self.right.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.left.__exit__(exc_type, exc_val, exc_tb)
        finally:
            self.right.__exit__(exc_type, exc_val, exc_tb)


#
# Concrete environment classes
#


class DictEnv(Env):
    """
    A simple dictionary environment; more-or-less boring!
    """

    def __init__(self, d=None):
        self.d = d or {}

    def bind(self, name, value):
        return DictEnv(dict(list(self.d.items()) + [(name, value)]))

    def lookup(self, name):
        try:
            return self.d[name]
        except KeyError:
            raise NotFound(unwrap_val(name))

    def replace(self, data):
        if isinstance(data, dict):
            return DictEnv(data)
        else:
            raise CannotReplace()


class JsonPathEnv(Env):
    """
    An environment like those that map names to variables, but
    supporting dereferencing an JsonPath expression. Note that it never
    fails a lookup, but always returns an empty  list.

    It also interns all parsed expressions
    """

    def __init__(self, bindings=None):
        self.__bindings = bindings or {}
        self.__restrict_to_root = bool(
            jsonpath.Fields("__root_only").find(self.__bindings)
        )

        # Currently hardcoded because it is a global is jsonpath-ng
        # Probably not widely used, but will require refactor if so
        jsonpath.auto_id_field = "id"

    def parse(self, jsonpath_string):
        if jsonpath_string not in JSONPATH_CACHE:
            JSONPATH_CACHE[jsonpath_string] = parse_jsonpath(jsonpath_string)
        return JSONPATH_CACHE[jsonpath_string]

    def lookup(
        self,
        name: Union[str, jsonpath.JSONPath]
    ) -> RepeatableIterator:
        if isinstance(name, str):
            jsonpath_expr = self.parse(name)
        elif isinstance(name, jsonpath.JSONPath):
            jsonpath_expr = name
        else:
            raise NotFound(unwrap_val(name))

        # special case for 'id'
        if self.__restrict_to_root and str(jsonpath_expr) != 'id':
            expr, _ = split_leftmost(jsonpath_expr)
            if not isinstance(expr, jsonpath.Root):
                return RepeatableIterator(lambda: iter(()))

        def iterator(jsonpath_expr=jsonpath_expr):  # Capture closure
            for datum in jsonpath_expr.find(self.__bindings):
                # HACK: The auto id from jsonpath_ng is good, but we
                # lose it when we do .value here, so just slap it on if
                # not present
                if isinstance(datum.value, dict) and 'id' not in datum.value:
                    datum.value['id'] = jsonpath.AutoIdForDatum(datum).value
                yield datum

        return RepeatableIterator(iterator)

    @overload
    def bind(self, key: str, value: Any, *args) -> Env:
        ...

    @overload
    def bind(self, bindings: Dict[str, Any], *args) -> Env:
        ...

    def bind(self, *args):
        new_bindings = dict(self.__bindings)
        if isinstance(args[0], dict):
            new_bindings.update(args[0])
            return self.__class__(new_bindings)

        elif isinstance(args[0], str):
            new_bindings[args[0]] = args[1]
            return self.__class__(new_bindings)

        else:
            raise ValueError('Bad args to Env.bind')

    def replace(self, data):
        return self.__class__(data)


#
# Actual concrete environments, basically with built-in functions.
#


def _not_val(val):
    return val is None or val == []


def _to_unicode(val):
    if isinstance(val, bytes):
        return val.decode('utf8')
    elif not isinstance(val, str):
        return str(val)

    return val


@unwrap('val')
def str2bool(val):
    if _not_val(val):
        return False

    if isinstance(val, bool):
        return val

    val = _to_unicode(val)
    return bool(val) and val.lower() in {'true', 't', '1'}


@unwrap('val')
def str2num(val):
    if _not_val(val):
        return None

    try:
        return int(val)
    except ValueError:
        try:
            return float(val)
        except ValueError:
            return None


@unwrap('val')
def str2date(val):
    import dateutil.parser as parser
    if not val:
        return None

    val = _to_unicode(val)

    try:
        date = parser.parse(val)
    except ValueError:
        return

    if date.tzinfo is not None:
        try:
            date = date.astimezone(pytz.utc)
        except ValueError:
            pass

    return date.replace(microsecond=0, tzinfo=None)


@unwrap('val')
def bool2int(val):
    return int(str2bool(val))


@unwrap('val')
def sha1(val):
    if _not_val(val):
        return None

    if not isinstance(val, bytes):
        val = str(val).encode('utf8')

    return hashlib.sha1(val).hexdigest()


@unwrap('val')
def selected_at(val, index):
    if not val:
        return None

    try:
        index = int(index)
    except ValueError:
        return "Error: index must be an integer"

    val = _to_unicode(val)

    try:
        return val.split()[index]
    except (IndexError, ValueError):
        return None


@unwrap('val')
def selected(val, reference):
    if not val:
        return None

    val = _to_unicode(val)

    parts = val.split()
    return reference in parts


@unwrap('val')
def count_selected(val):
    if not val:
        return None

    val = _to_unicode(val)

    return len(val.split())


@unwrap('val')
def json2str(val):
    if isinstance(val, str):
        return val
    try:
        return json.dumps(val)
    except ValueError:
        return


@unwrap('val')
def format_uuid(val):
    """
    Renders a hex UUID in hyphen-separated groups

    >>> format_uuid('00a3e0194ce1458794c50971dee2de22')
    '00a3e019-4ce1-4587-94c5-0971dee2de22'
    >>> format_uuid(0x00a3e0194ce1458794c50971dee2de22)
    '00a3e019-4ce1-4587-94c5-0971dee2de22'
    """
    if not val:
        return None
    if isinstance(val, int):
        val = hex(val)
    try:
        return str(uuid.UUID(val))
    except ValueError:
        return None


def join(*args):
    args_ = [unwrap_val(arg) for arg in args]
    try:
        return args_[0].join(args_[1:])
    except TypeError:
        return '""'


@unwrap('val')
def default(val, default_val):
    if not val:
        return default_val
    return val


@unwrap('val')
def attachment_url(val):
    if not val:
        return None
    from commcare_export.minilinq import Apply, Reference, Literal
    return Apply(
        Reference('template'),
        Literal('{}/a/{}/api/form/attachment/{}/{}'),
        Reference('commcarehq_base_url'),
        Reference('$.domain'),
        Reference('$.id'),
        Literal(val)
    )


@unwrap('val')
def form_url(val):
    return _doc_url('form_data')


@unwrap('val')
def case_url(val):
    return _doc_url('case_data')


def _doc_url(url_path):
    from commcare_export.minilinq import Apply, Reference, Literal
    return Apply(
        Reference('template'),
        Literal('{}/a/{}/reports/' + url_path + '/{}/'),
        Reference('commcarehq_base_url'),
        Reference('$.domain'),
        Reference('$.id'),
    )


def template(format_template, *args):
    args_ = [unwrap_val(arg) for arg in args]
    return format_template.format(*args_)


def _or(*args):
    return _or_impl(unwrap_val, *args)


def _or_raw(*args):

    def unwrap_iter(arg):
        if isinstance(arg, RepeatableIterator):
            return list(arg)
        return arg

    return _or_impl(unwrap_iter, *args)


def _or_impl(_unwrap, *args):
    unwrapped_args = (_unwrap(arg) for arg in args)
    vals = (val for val in unwrapped_args if val is not None and val != [])
    try:
        return next(vals)
    except StopIteration:
        pass


@unwrap('val')
def substr(val, start, end):
    if not val:
        return None

    if start < 0 or end < 0:
        return None

    val = _to_unicode(val)

    return val[start:end]


@unwrap('val')
def unique(val):
    if isinstance(val, list):
        return list(set(val))
    return val


class BuiltInEnv(DictEnv):
    """
    A built-in environment of operators and functions which does not
    support replacement or bindings.

    For convenience, this environment has been chosen to queue up tables
    to be written out, since it will be the first env involved in almost
    any situation.
    """

    def __init__(self, d=None):
        self.__tables = []
        d = d or {}
        d.update({
            '+': operator.__add__,
            '-': operator.__sub__,
            '*': operator.__mul__,
            '//': operator.__floordiv__,
            '/': operator.__truediv__,
            '>': operator.__gt__,
            '>=': operator.__ge__,
            '<': operator.__lt__,
            '<=': operator.__le__,
            'len': len,
            'bool': bool,
            'str2bool': str2bool,
            'bool2int': bool2int,
            'str2num': str2num,
            'str2date': str2date,
            'json2str': json2str,
            'format-uuid': format_uuid,
            'selected': selected,
            'selected-at': selected_at,
            'count-selected': count_selected,
            'join': join,
            'default': default,
            'template': template,
            'form_url': form_url,
            'case_url': case_url,
            'attachment_url': attachment_url,
            'filter_empty': _not_val,
            'or': _or,
            'sha1': sha1,
            'substr': substr,
            '_or_raw': _or_raw,  # for internal use,
            'unique': unique
        })
        super(BuiltInEnv, self).__init__(d)

    def bind(self, name, value):
        raise CannotBind()

    def replace(self, data):
        raise CannotReplace()


class EmitterEnv(Env):

    def __init__(self, writer):
        self.writer = writer
        self.emitted = False

    def __enter__(self):
        self.writer.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.writer.__exit__(exc_type, exc_val, exc_tb)

    def bind(self, name, value):
        raise CannotBind()

    def replace(self, data):
        raise CannotReplace()

    def lookup(self, key):
        raise NotFound()

    def emit_table(self, table_spec):
        self.emitted = True
        table_spec.rows = self._unwrap_row_vals(table_spec.rows)
        self.writer.write_table(table_spec)

    def has_emitted_tables(self):
        return self.emitted

    @staticmethod
    def _unwrap_row_vals(rows):
        """
        The XMLtoJSON conversion in CommCare can result in a field being
        a JSON object instead of a simple field (if the XML tag has
        attributes or different namespace from the default). In this
        case the actual value of the XML element is stored in a '#text'
        field.
        """

        def _unwrap_val(val):
            if isinstance(val, dict):
                if '#text' in val:
                    return val.get('#text')
                elif all(key == 'id' or key.startswith('@') for key in val):
                    # this implies the XML element was empty since all
                    # keys are from attributes
                    return ''
            return val

        for row in rows:
            yield [_unwrap_val(val) for val in row]



from __future__ import (
    absolute_import,
    division,
    generators,
    nested_scopes,
    print_function,
    unicode_literals,
)

import copy
import logging
import time
from collections import OrderedDict
from math import ceil
from urllib.parse import urlencode

import backoff
import requests
from requests.auth import AuthBase, HTTPDigestAuth

import commcare_export
from commcare_export.repeatable_iterator import RepeatableIterator
from datetime import datetime

AUTH_MODE_PASSWORD = 'password'
AUTH_MODE_APIKEY = 'apikey'

logger = logging.getLogger(__name__)

LATEST_KNOWN_VERSION = '0.5'
RESOURCE_REPEAT_LIMIT = 10

def on_wait(details):
    time_to_wait = details["wait"]
    logger.warning(f"Rate limit reached. Waiting for {time_to_wait} seconds.")

def on_backoff(details):
    _log_backoff(details, 'Waiting for retry.')


def on_giveup(details):
    _log_backoff(details, 'Giving up.')


def _log_backoff(details, action_message):
    details['__suffix'] = action_message
    logger.warning(
        "Request failed after {tries} attempts ({elapsed:.1f}s). {__suffix}"
        .format(**details)
    )


def is_client_error(ex):
    logger.info(str(ex))
    if hasattr(ex, 'response') and ex.response is not None:
        if ex.response.status_code == 429:
            # "Too Many Requests": HQ wants us to back off
            return False
        return 400 <= ex.response.status_code < 500
    return False


class ResourceRepeatException(Exception):

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class CommCareHqClient(object):
    """
    A connection to CommCareHQ for a particular version, project, and user.
    """

    def __init__(
        self,
        url,
        project,
        username,
        password,
        auth_mode=AUTH_MODE_PASSWORD,
        version=LATEST_KNOWN_VERSION,
        checkpoint_manager=None
    ):
        self.version = version
        self.url = url
        self.project = project
        self.__auth = self._get_auth(username, password, auth_mode)
        self.__session = None

    def _get_auth(self, username, password, mode):
        if mode == AUTH_MODE_PASSWORD:
            return HTTPDigestAuth(username, password)
        elif mode == AUTH_MODE_APIKEY:
            return ApiKeyAuth(username, password)
        else:
            raise Exception('Unknown auth mode: %s' % mode)

    @property
    def session(self):
        if self.__session == None:
            self.__session = requests.Session()
            self.__session.headers.update({
                'User-Agent': f'commcare-export/{commcare_export.__version__}'
            })
        return self.__session

    @session.setter
    def session(self, session):
        """Used for overriding the session in unit tests"""
        self.__session = session

    @property
    def api_url(self):
        return '%s/a/%s/api/v%s' % (self.url, self.project, self.version)

    def _should_raise_for_status(self, response):
        return "Retry-After" not in response.headers

    def get(self, resource, params=None):
        """
        Gets the named resource. When the server returns a 429 (too many requests), the process will sleep for
        the amount of seconds specified in the Retry-After header from the response, after which it will raise
        an exception to trigger the retry action.

        Currently a bit of a vulnerable stub that works for this
        particular use case in the hands of a trusted user; would likely
        want this to work like (or via) slumber.
        """
        @backoff.on_predicate(
            backoff.runtime,
            predicate=lambda r: r.status_code == 429,
            value=lambda r: ceil(float(r.headers.get("Retry-After", 1.0))),
            jitter=None,
            on_backoff=on_wait,
        )
        @backoff.on_exception(
            backoff.expo,
            requests.exceptions.RequestException,
            max_time=300,
            giveup=is_client_error,
            on_backoff=on_backoff,
            on_giveup=on_giveup
        )
        def _get(resource, params=None):
            logger.debug("Fetching '%s' batch: %s", resource, params)
            resource_url = f'{self.api_url}/{resource}/'
            response = self.session.get(
                resource_url, params=params, auth=self.__auth, timeout=60
            )
            if self._should_raise_for_status(response):
                response.raise_for_status()
            return response

        response = _get(resource, params)
        return response.json()

    def iterate(
        self,
        resource,
        paginator,
        params=None,
        checkpoint_manager=None,
    ):
        """
        Assumes the endpoint is a list endpoint, and iterates over it
        making a lot of assumptions that it is like a tastypie endpoint.
        """
        UNKNOWN_COUNT = 'unknown'
        params = dict(params or {})

        def iterate_resource(resource=resource, params=params):
            more_to_fetch = True
            last_batch_ids = set()
            total_count = UNKNOWN_COUNT
            fetched = 0
            repeat_counter = 0
            last_params = None

            while more_to_fetch:
                if params == last_params:
                    repeat_counter += 1
                else:
                    repeat_counter = 0
                if repeat_counter >= RESOURCE_REPEAT_LIMIT:
                    raise ResourceRepeatException(
                        f"Requested resource '{resource}' {repeat_counter} "
                        "times with same parameters"
                    )

                batch = self.get(resource, params)
                last_params = copy.copy(params)
                batch_meta = batch['meta']
                if total_count == UNKNOWN_COUNT or fetched >= total_count:
                    if batch_meta.get('total_count'):
                        total_count = int(batch_meta['total_count'])
                    else:
                        total_count = UNKNOWN_COUNT
                    fetched = 0

                batch_objects = batch['objects']
                fetched += len(batch_objects)
                logger.debug('Received %s of %s', fetched, total_count)
                if not batch_objects:
                    more_to_fetch = False
                else:
                    got_new_data = False
                    for obj in batch_objects:
                        if obj['id'] not in last_batch_ids:
                            yield obj
                            got_new_data = True

                    if batch_meta.get('next'):
                        last_batch_ids = {obj['id'] for obj in batch_objects}
                        params = paginator.next_page_params_from_batch(batch)
                        if not params:
                            more_to_fetch = False
                    else:
                        more_to_fetch = False

                    limit = batch_meta.get('limit')
                    if more_to_fetch:
                        # Handle the case where API is 'non-counting'
                        # and repeats the last batch
                        repeated_last_page_of_non_counting_resource = (
                            not got_new_data and total_count == UNKNOWN_COUNT
                            and (limit and len(batch_objects) < limit)
                        )
                        more_to_fetch = not repeated_last_page_of_non_counting_resource

                    paginator.set_checkpoint(
                        checkpoint_manager,
                        batch,
                        not more_to_fetch
                    )

        return RepeatableIterator(iterate_resource)


class MockCommCareHqClient(object):
    """
    An in-memory mock of the hq client, instantiated with a simple
    mapping of resource and params to results.

    Since dictionaries are not hashable, the mapping is written as a
    pair of tuples, handled appropriately internally.

    MockCommCareHqClient({
        'forms': [
            (
                {'_search': 'test1'},
                [
                   ... objects ...
                ]
            ),
        ]
    })
    """

    def __init__(self, mock_data):
        self.mock_data = {
            resource: {
                _params_to_url(params): result
                for (params, result) in resource_results
            } for (resource, resource_results) in mock_data.items()
        }

    def iterate(
        self, resource, paginator, params=None, checkpoint_manager=None
    ):
        logger.debug(
            'Mock client call to resource "%s" with params "%s"',
            resource,
            params
        )
        return self.mock_data[resource][_params_to_url(params)]

    def get(self, resource):
        logger.debug('Mock client call to get resource "%s"', resource)
        objects = self.mock_data[resource][_params_to_url({'get': True})]
        if objects:
            return {
                'meta': {
                    'limit': len(objects),
                    'next': None,
                    'offset': 0,
                    'previous': None,
                    'total_count': len(objects)
                },
                'objects': objects
            }
        else:
            return None


def _params_to_url(params):
    return urlencode(OrderedDict(sorted(params.items())))


class ApiKeyAuth(AuthBase):

    def __init__(self, username, apikey):
        self.username = username
        self.apikey = apikey

    def __eq__(self, other):
        return all([
            self.username == getattr(other, 'username', None),
            self.apikey == getattr(other, 'apikey', None)
        ])

    def __hash__(self):
        return hash((self.username, self.apikey))

    def __ne__(self, other):
        return not self == other

    def __call__(self, r):
        r.headers['Authorization'] = f'apikey {self.username}:{self.apikey}'
        return r



class DataExportException(Exception):
    message = None


class LongFieldsException(DataExportException):

    def __init__(self, long_fields, max_length):
        self.long_fields = long_fields
        self.max_length = max_length

    @property
    def message(self):
        message = ''
        for table, headers in self.long_fields.items():
            message += (
                f'Table "{table}" has field names longer than the maximum '
                f'allowed for this database ({self.max_length}):\n'
            )
            for header in headers:
                message += '    {}\n'.format(header)

        message += (
            '\nPlease adjust field names to be within the maximum length '
            f'limit of {self.max_length}'
        )
        return message


class MissingColumnException(DataExportException):

    def __init__(self, errors_by_sheet):
        self.errors_by_sheet = errors_by_sheet

    @property
    def message(self):
        lines = [
            'Sheet "{}" is missing definitions for required fields: "{}"'
            .format(sheet, '", "'.join(missing_cols))
            for (sheet, missing_cols) in self.errors_by_sheet.items()
        ]
        return '\n'.join(lines)


class MissingQueryFileException(DataExportException):

    def __init__(self, query_file):
        self.query_file = query_file

    @property
    def message(self):
        return f'Query file not found: {self.query_file}'


class ReservedTableNameException(DataExportException):

    def __init__(self, conflicting_name):
        self.conflicting_name = conflicting_name

    @property
    def message(self):
        return (
            f'Table name "{self.conflicting_name}" conflicts with an internal '
            f'table name. Please export to a different table.'
        )



"""Add detail to checkpoint

Revision ID: a56c82a8d02e
Revises: f4fd4c80f40a
Create Date: 2021-01-22 16:35:07.063082

"""
from alembic import op
import sqlalchemy as sa


revision = 'a56c82a8d02e'
down_revision = 'f4fd4c80f40a'
branch_labels = None
depends_on = None


def upgrade():
    url = op.get_bind().engine.url
    collation = 'utf8_bin' if 'mysql' in url.drivername else None
    op.add_column(
        'commcare_export_runs',
        sa.Column('data_source', sa.Unicode(255, collation=collation))
    )
    op.add_column(
        'commcare_export_runs',
        sa.Column('last_doc_id', sa.Unicode(255, collation=collation))
    )

def downgrade():
    op.drop_column('commcare_export_runs', 'data_source')
    op.drop_column('commcare_export_runs', 'last_doc_id')



import csv
import os
import re
import unittest
from argparse import Namespace
from copy import copy
from itertools import zip_longest
from unittest import mock

import sqlalchemy
from tests.utils import SqlWriterWithTearDown

import pytest
from commcare_export.checkpoint import (
    Checkpoint,
    CheckpointManager,
    session_scope,
)
from commcare_export.cli import CLI_ARGS, main_with_args
from commcare_export.commcare_hq_client import (
    CommCareHqClient,
    MockCommCareHqClient,
    _params_to_url,
)
from commcare_export.commcare_minilinq import PaginationMode
from commcare_export.specs import TableSpec
from commcare_export.writers import JValueTableWriter

CLI_ARGS_BY_NAME = {arg.name: arg for arg in CLI_ARGS}

DEFAULT_BATCH_SIZE = 200


def make_args(project='test', username='test', password='test', **kwargs):
    kwargs['project'] = project
    kwargs['username'] = username
    kwargs['password'] = password

    args_by_name = copy(CLI_ARGS_BY_NAME)
    namespace = Namespace()
    for name, val in kwargs.items():
        args_by_name.pop(name)
        setattr(namespace, name, val)

    for name, arg in args_by_name.items():
        setattr(namespace, name, arg.default)

    return namespace


def mock_hq_client(include_parent):
    return MockCommCareHqClient({
        'form': [({
            'limit': DEFAULT_BATCH_SIZE,
            'order_by': 'indexed_on'
        }, [
            {
                'id': 1,
                'form': {
                    'name': 'f1',
                    'case': {
                        '@case_id': 'c1'
                    }
                },
                'metadata': {
                    'userID': 'id1'
                }
            },
            {
                'id': 2,
                'form': {
                    'name': 'f2',
                    'case': {
                        '@case_id': 'c2'
                    }
                },
                'metadata': {
                    'userID': 'id2'
                }
            },
        ]),],
        'case': [({
            'limit': DEFAULT_BATCH_SIZE,
            'order_by': 'indexed_on'
        }, [
            {
                'id': 'case1'
            },
            {
                'id': 'case2'
            },
        ])],
        'user': [({
            'limit': DEFAULT_BATCH_SIZE
        }, [{
            'id': 'id1',
            'email': 'em1',
            'first_name': 'fn1',
            'last_name': 'ln1',
            'user_data': {
                'commcare_location_id': 'lid1',
                'commcare_location_ids': ['lid1', 'lid2'],
                'commcare_project': 'p1'
            },
            'username': 'u1'
        }, {
            'id': 'id2',
            'default_phone_number': 'pn2',
            'email': 'em2',
            'first_name': 'fn2',
            'last_name': 'ln2',
            'resource_uri': 'ru0',
            'user_data': {
                'commcare_location_id': 'lid2',
                'commcare_project': 'p2'
            },
            'username': 'u2'
        }])],
        'location_type': [({
            'limit': DEFAULT_BATCH_SIZE
        }, [{
            'administrative': True,
            'code': 'hq',
            'domain': 'd1',
            'id': 1,
            'name': 'HQ',
            'parent': None,
            'resource_uri': 'lt1',
            'shares_cases': False,
            'view_descendants': True
        }, {
            'administrative': False,
            'code': 'local',
            'domain': 'd1',
            'id': 2,
            'name': 'Local',
            'parent': 'lt1',
            'resource_uri': 'lt2',
            'shares_cases': True,
            'view_descendants': True
        }])],
        'location': [({
            'limit': DEFAULT_BATCH_SIZE
        }, [{
            'id': 'id1',
            'created_at': '2020-04-01T21:57:26.403053',
            'domain': 'd1',
            'external_id': 'eid1',
            'last_modified': '2020-04-01T21:58:23.88343',
            'latitude': '11.2',
            'location_data': {
                'p1': 'ld1'
            },
            'location_id': 'lid1',
            'location_type': 'lt1',
            'longitude': '-20.5',
            'name': 'n1',
            'resource_uri': 'ru1',
            'site_code': 'sc1'
        }, {
            'id': 'id2',
            'created_at': '2020-04-01T21:58:47.627371',
            'domain': 'd2',
            'last_modified': '2020-04-01T21:59:16.018411',
            'latitude': '-56.3',
            'location_data': {
                'p1': 'ld2'
            },
            'location_id': 'lid2',
            'location_type': 'lt2',
            'longitude': '18.7',
            'name': 'n2',
            'parent': 'ru1' if include_parent else None,
            'resource_uri': 'ru2',
            'site_code': 'sc2'
        }])],
    })


EXPECTED_MULTIPLE_TABLES_RESULTS = [{
    "name": "Forms",
    "headings": ["id", "name"],
    "rows": [["1", "f1"], ["2", "f2"]],
}, {
    "name": "Other cases",
    "headings": ["id"],
    "rows": [["case1"], ["case2"]],
}, {
    "name": "Cases",
    "headings": ["case_id"],
    "rows": [["c1"], ["c2"]],
}]

EXPECTED_USERS_RESULTS = [{
    "name":
        "commcare_users",
    "headings": [
        "id", "default_phone_number", "email", "first_name", "groups",
        "last_name", "phone_numbers", "resource_uri", "commcare_location_id",
        "commcare_location_ids", "commcare_primary_case_sharing_id",
        "commcare_project", "username"
    ],
    "rows": [[
        "id1", None, "em1", "fn1", None, "ln1", None, None, "lid1",
        "lid1,lid2", None, "p1", "u1"
    ],
             [
                 "id2", "pn2", "em2", "fn2", None, "ln2", None, "ru0", "lid2",
                 None, None, "p2", "u2"
             ]]
}]


def get_expected_locations_results(include_parent):
    return [{
        "name":
            "commcare_locations",
        "headings": [
            "id", "created_at", "domain", "external_id", "last_modified",
            "latitude", "location_data", "location_id", "location_type",
            "longitude", "name", "parent", "resource_uri", "site_code",
            "location_type_administrative", "location_type_code",
            "location_type_name", "location_type_parent", "local", "hq"
        ],
        "rows": [[
            "id1", "2020-04-01 21:57:26", "d1", "eid1", "2020-04-01 21:58:23",
            "11.2", '{"p1": "ld1", "id": "id1.location_data"}', "lid1", "lt1",
            "-20.5", "n1", None, "ru1", "sc1", True, "hq", "HQ", None, None,
            "lid1"
        ],
                 [
                     "id2", "2020-04-01 21:58:47", "d2", None,
                     "2020-04-01 21:59:16", "-56.3",
                     '{"p1": "ld2", "id": "id2.location_data"}',
                     "lid2", "lt2", "18.7", "n2",
                     ("ru1" if include_parent else None), "ru2", "sc2", False,
                     "local", "Local", "lt1", "lid2",
                     ("lid1" if include_parent else None)
                 ]]
    }]


class TestCli(unittest.TestCase):

    def _test_cli(self, args, expected):
        writer = JValueTableWriter()
        with mock.patch(
            'commcare_export.cli._get_writer', return_value=writer
        ):
            main_with_args(args)

        for table in expected:
            assert writer.tables[table['name']] == TableSpec(**table)

    @mock.patch(
        'commcare_export.cli._get_api_client',
        return_value=mock_hq_client(True)
    )
    def test_cli(self, mock_client):
        args = make_args(
            query='tests/008_multiple-tables.xlsx', output_format='json'
        )
        self._test_cli(args, EXPECTED_MULTIPLE_TABLES_RESULTS)

    @mock.patch(
        'commcare_export.cli._get_api_client',
        return_value=mock_hq_client(True)
    )
    def test_cli_just_users(self, mock_client):
        args = make_args(output_format='json', users=True)
        self._test_cli(args, EXPECTED_USERS_RESULTS)

    @mock.patch(
        'commcare_export.cli._get_api_client',
        return_value=mock_hq_client(True)
    )
    def test_cli_table_plus_users(self, mock_client):
        args = make_args(
            query='tests/008_multiple-tables.xlsx',
            output_format='json',
            users=True
        )
        self._test_cli(
            args, EXPECTED_MULTIPLE_TABLES_RESULTS + EXPECTED_USERS_RESULTS
        )

    @mock.patch(
        'commcare_export.cli._get_api_client',
        return_value=mock_hq_client(True)
    )
    def test_cli_just_locations(self, mock_client):
        args = make_args(output_format='json', locations=True)
        self._test_cli(args, get_expected_locations_results(True))

    @mock.patch(
        'commcare_export.cli._get_api_client',
        return_value=mock_hq_client(False)
    )
    def test_cli_locations_without_parents(self, mock_client):
        args = make_args(output_format='json', locations=True)
        self._test_cli(args, get_expected_locations_results(False))

    @mock.patch(
        'commcare_export.cli._get_api_client',
        return_value=mock_hq_client(True)
    )
    def test_cli_table_plus_locations(self, mock_client):
        args = make_args(
            query='tests/008_multiple-tables.xlsx',
            output_format='json',
            locations=True
        )
        self._test_cli(
            args, EXPECTED_MULTIPLE_TABLES_RESULTS
            + get_expected_locations_results(True)
        )


@pytest.fixture(scope='function')
def writer(pg_db_params):
    writer = SqlWriterWithTearDown(
        pg_db_params['url'], poolclass=sqlalchemy.pool.NullPool
    )
    yield writer
    writer.tear_down()


@pytest.fixture(scope='function')
def checkpoint_manager(pg_db_params):
    cm = CheckpointManager(
        pg_db_params['url'],
        'query',
        '123',
        'test',
        'hq',
        poolclass=sqlalchemy.pool.NullPool
    )
    cm.create_checkpoint_table()
    return cm


def _pull_data(writer, checkpoint_manager, query, since, until, batch_size=10):
    args = make_args(
        query=query,
        output_format='sql',
        output='',
        username=os.environ['HQ_USERNAME'],
        password=os.environ['HQ_API_KEY'],
        auth_mode='apikey',
        project='corpora',
        batch_size=batch_size,
        since=since,
        until=until,
    )

    # set this so that it gets written to the checkpoints
    checkpoint_manager.query = query

    # have to mock these to override the pool class otherwise they hold
    # the db connection open
    writer_patch = mock.patch(
        'commcare_export.cli._get_writer', return_value=writer
    )
    checkpoint_patch = mock.patch(
        'commcare_export.cli._get_checkpoint_manager',
        return_value=checkpoint_manager
    )
    with writer_patch, checkpoint_patch:
        main_with_args(args)


def _check_data(writer, expected, table_name, columns):
    actual = [
        list(row) for row in writer.engine
        .execute(f'SELECT {", ".join(columns)} FROM "{table_name}"')
    ]

    message = ''
    if actual != expected:
        message += 'Data not equal to expected:\n'
        if len(actual) != len(expected):
            message += '    {} rows compared to {} expected\n'.format(
                len(actual), len(expected)
            )
        message += 'Diff:\n'
        for i, rows in enumerate(zip_longest(actual, expected)):
            if rows[0] != rows[1]:
                message += '{}: {} != {}\n'.format(i, rows[0], rows[1])
        assert actual == expected, message


@pytest.mark.dbtest
class TestCLIIntegrationTests(object):

    def test_write_to_sql_with_checkpoints(
        self, writer, checkpoint_manager, caplog
    ):
        with open('tests/009_expected_form_data.csv', 'r') as f:
            reader = csv.reader(f)
            expected_form_data = list(reader)[1:]

        _pull_data(
            writer, checkpoint_manager, 'tests/009_integration.xlsx',
            '2012-01-01', '2017-08-29'
        )
        self._check_checkpoints(caplog, ['forms', 'batch', 'final'])
        self._check_data(writer, expected_form_data[:13], 'forms')

        caplog.clear()
        _pull_data(
            writer,
            checkpoint_manager,
            'tests/009_integration.xlsx',
            None,
            '2020-10-11',
            batch_size=8
        )
        self._check_data(writer, expected_form_data, 'forms')
        self._check_checkpoints(caplog, ['forms', 'batch', 'final'])

        runs = list(
            writer.engine.execute(
                'SELECT * FROM commcare_export_runs '
                'WHERE query_file_name = %s',

                'tests/009_integration.xlsx'
            )
        )
        assert len(runs) == 2, runs

    def test_write_to_sql_with_checkpoints_multiple_tables(
        self, writer, checkpoint_manager, caplog
    ):
        with open('tests/009b_expected_form_1_data.csv', 'r') as f:
            reader = csv.reader(f)
            expected_form_1_data = list(reader)[1:]

        with open('tests/009b_expected_form_2_data.csv', 'r') as f:
            reader = csv.reader(f)
            expected_form_2_data = list(reader)[1:]

        _pull_data(
            writer, checkpoint_manager, 'tests/009b_integration_multiple.xlsx',
            None, '2020-10-11'
        )
        self._check_checkpoints(
            caplog, ['forms_1', 'batch', 'batch', 'final', 'forms_2', 'final']
        )
        self._check_checkpoints(
            caplog,
            ['forms_1', 'forms_1', 'forms_1', 'forms_1', 'forms_2', 'forms_2']
        )
        self._check_data(writer, expected_form_1_data, 'forms_1')
        self._check_data(writer, expected_form_2_data, 'forms_2')

        runs = list(
            writer.engine.execute(
                'SELECT table_name, since_param '
                'FROM commcare_export_runs '
                'WHERE query_file_name = %s',

                'tests/009b_integration_multiple.xlsx'
            )
        )
        assert {r[0]: r[1] for r in runs} == {
            'forms_1': '2017-09-02T20:05:35.459547',
            'forms_2': '2020-06-01T17:43:26.107701',
        }

    def _check_data(self, writer, expected, table_name):
        _check_data(writer, expected, table_name, ['id', 'name', 'indexed_on'])

    def _check_checkpoints(self, caplog, expected):
        # Depends on the logging in the CheckpointManager._set_checkpoint
        # method
        log_messages = [
            record[2]
            for record in caplog.record_tuples
            if record[0] == 'commcare_export.checkpoint'
        ]
        fail = False
        message = ''
        for i, items in enumerate(zip_longest(expected, log_messages)):
            if not items[0] or not items[1] or items[0] not in items[1]:
                message += 'X {}: {} not in {}\n'.format(i, items[0], items[1])
                fail = True
            else:
                message += '✓ {}: {} in {}\n'.format(i, items[0], items[1])
        assert not fail, 'Checkpoint comparison failed:\n' + message


# Conflicting types for 'count' will cause errors when inserting into
# database.
CONFLICTING_TYPES_CLIENT = MockCommCareHqClient({
    'case': [({
        'limit': DEFAULT_BATCH_SIZE,
        'order_by': 'indexed_on'
    }, [{
        'id': 1,
        'name': 'n1',
        'count': 10
    }, {
        'id': 2,
        'name': 'f2',
        'count': 'abc'
    }]),],
})


class MockCheckpointingClient(CommCareHqClient):
    """
    Mock client that uses the main client for iteration but overrides
    the data request to return mocked data.

    Note this client needs to be re-initialized after use.
    """

    def __init__(self, mock_data):
        self.mock_data = {
            resource: {
                _params_to_url(params): result
                for params, result in resource_results
            } for resource, resource_results in mock_data.items()
        }
        self.totals = {
            resource: sum(len(results) for _, results in resource_results
                         ) for resource, resource_results in mock_data.items()
        }

    def get(self, resource, params=None):
        mock_requests = self.mock_data[resource]
        key = _params_to_url(params)
        objects = mock_requests.pop(key)
        if objects:
            return {
                'meta': {
                    'limit': len(objects),
                    'next': bool(mock_requests),
                    'offset': 0,
                    'previous': None,
                    'total_count': self.totals[resource]
                },
                'objects': objects
            }
        else:
            return None


def get_conflicting_types_checkpoint_client():
    return MockCheckpointingClient({
        'case': [
            ({
                'limit': DEFAULT_BATCH_SIZE,
                'order_by': 'indexed_on'
            }, [{
                'id': "doc 1",
                'name': 'n1',
                'count': 10,
                'indexed_on': '2012-04-23T05:13:01.000000Z'
            }, {
                'id': "doc 2",
                'name': 'f2',
                'count': 123,
                'indexed_on': '2012-04-24T05:13:01.000000Z'
            }]),
            ({
                'limit': DEFAULT_BATCH_SIZE,
                'order_by': 'indexed_on',
                'indexed_on_start': '2012-04-24T05:13:01'
            }, [{
                'id': "doc 3",
                'name': 'n1',
                'count': 10,
                'indexed_on': '2012-04-25T05:13:01.000000Z'
            }, {
                'id': "doc 4",
                'name': 'f2',
                'count': 'abc',
                'indexed_on': '2012-04-26T05:13:01.000000Z'
            }]),
        ],
    })


@pytest.fixture(scope='function')
def strict_writer(db_params):
    writer = SqlWriterWithTearDown(
        db_params['url'],
        poolclass=sqlalchemy.pool.NullPool,
        strict_types=True
    )
    yield writer
    writer.tear_down()


@pytest.fixture(scope='function')
def all_db_checkpoint_manager(db_params):
    cm = CheckpointManager(
        db_params['url'],
        'query',
        '123',
        'test',
        'hq',
        poolclass=sqlalchemy.pool.NullPool
    )
    cm.create_checkpoint_table()
    yield cm
    with session_scope(cm.Session) as session:
        session.query(Checkpoint).delete(synchronize_session='fetch')


def _pull_mock_data(
    writer,
    checkpoint_manager,
    api_client,
    query,
    start_over=None,
    since=None
):
    args = make_args(
        query=query,
        output_format='sql',
        start_over=start_over,
        since=since,
    )

    assert not (checkpoint_manager and since), \
        "'checkpoint_manager' must be None when using 'since'"

    if checkpoint_manager:
        # set this so that it gets written to the checkpoints
        checkpoint_manager.query = query

    # have to mock these to override the pool class otherwise they hold
    # the db connection open
    api_client_patch = mock.patch(
        'commcare_export.cli._get_api_client', return_value=api_client
    )
    writer_patch = mock.patch(
        'commcare_export.cli._get_writer', return_value=writer
    )
    checkpoint_patch = mock.patch(
        'commcare_export.cli._get_checkpoint_manager',
        return_value=checkpoint_manager
    )
    with api_client_patch, writer_patch, checkpoint_patch:
        return main_with_args(args)


@pytest.mark.dbtest
class TestCLIWithDatabaseErrors(object):

    def test_cli_database_error(
        self, strict_writer, all_db_checkpoint_manager, capfd
    ):
        _pull_mock_data(
            strict_writer, all_db_checkpoint_manager, CONFLICTING_TYPES_CLIENT,
            'tests/013_ConflictingTypes.xlsx'
        )
        out, err = capfd.readouterr()

        expected_re = re.compile('Stopping because of database error')
        assert re.search(expected_re, out)

    def test_cli_database_error_checkpoint(
        self, strict_writer, all_db_checkpoint_manager, capfd
    ):
        _pull_mock_data(
            strict_writer, all_db_checkpoint_manager,
            get_conflicting_types_checkpoint_client(),
            'tests/013_ConflictingTypes.xlsx'
        )
        out, err = capfd.readouterr()

        expected_re = re.compile('Stopping because of database error')
        assert re.search(expected_re, out), out

        # expect checkpoint to have the date from the first batch and
        # not the 2nd
        runs = list(
            strict_writer.engine.execute(
                sqlalchemy.text(
                    'SELECT table_name, since_param, last_doc_id '
                    'FROM commcare_export_runs '
                    'WHERE query_file_name = :file'
                ),
                file='tests/013_ConflictingTypes.xlsx'
            )
        )
        assert runs == [
            ('Case', '2012-04-24T05:13:01', 'doc 2'),
        ]


# An input where missing fields should be added due to declared data
# types.
DATA_TYPES_CLIENT = MockCommCareHqClient({
    'form': [({
        'limit': DEFAULT_BATCH_SIZE,
        'order_by': 'indexed_on'
    }, [{
        'id': 1,
        'form': {}
    }, {
        'id': 2,
        'form': {}
    }]),],
})


@pytest.mark.dbtest
class TestCLIWithDataTypes(object):

    def test_cli_data_types_add_columns(
        self,
        writer,
        all_db_checkpoint_manager,
        capfd,
    ):
        _pull_mock_data(
            writer, all_db_checkpoint_manager, CONFLICTING_TYPES_CLIENT,
            'tests/014_ExportWithDataTypes.xlsx'
        )

        metadata = sqlalchemy.schema.MetaData(bind=writer.engine)
        table = sqlalchemy.Table(
            'forms',
            metadata,
            autoload_with=writer.engine,
        )
        cols = table.c
        assert sorted([c.name for c in cols]) == sorted([
            u'id', u'a_bool', u'an_int', u'a_date', u'a_datetime', u'a_text'
        ])

        # We intentionally don't check the types because SQLAlchemy
        # doesn't support type comparison, and even if we convert to
        # strings, the values are backend specific.

        values = [
            list(row) for row in writer.engine.execute('SELECT * FROM forms')
        ]

        assert values == [['1', None, None, None, None, None],
                          ['2', None, None, None, None, None]]


def get_indexed_on_client(page):
    p1 = MockCheckpointingClient({
        'case': [({
            'limit': DEFAULT_BATCH_SIZE,
            'order_by': 'indexed_on'
        }, [{
            'id': "doc 1",
            'name': 'n1',
            'indexed_on': '2012-04-23T05:13:01.000000Z'
        }, {
            'id': "doc 2",
            'name': 'n2',
            'indexed_on': '2012-04-24T05:13:01.000000Z'
        }])]
    })
    p2 = MockCheckpointingClient({
        'case': [({
            'limit': DEFAULT_BATCH_SIZE,
            'order_by': 'indexed_on',
            'indexed_on_start': '2012-04-24T05:13:01'
        }, [{
            'id': "doc 3",
            'name': 'n3',
            'indexed_on': '2012-04-25T05:13:01.000000Z'
        }, {
            'id': "doc 4",
            'name': 'n4',
            'indexed_on': '2012-04-26T05:13:01.000000Z'
        }])]
    })
    return [p1, p2][page]


@pytest.mark.dbtest
class TestCLIPaginationMode(object):

    def test_cli_pagination_fresh(self, writer, all_db_checkpoint_manager):
        checkpoint_manager = all_db_checkpoint_manager.for_dataset(
            "case", ["Case"]
        )

        _pull_mock_data(
            writer, all_db_checkpoint_manager, get_indexed_on_client(0),
            'tests/013_ConflictingTypes.xlsx'
        )
        self._check_data(writer, [["doc 1"], ["doc 2"]], "Case")
        self._check_checkpoint(
            checkpoint_manager, '2012-04-24T05:13:01', 'doc 2'
        )

        _pull_mock_data(
            writer, all_db_checkpoint_manager, get_indexed_on_client(1),
            'tests/013_ConflictingTypes.xlsx'
        )
        self._check_data(
            writer, [["doc 1"], ["doc 2"], ["doc 3"], ["doc 4"]], "Case"
        )
        self._check_checkpoint(
            checkpoint_manager, '2012-04-26T05:13:01', 'doc 4'
        )

    def test_cli_pagination_legacy(self, writer, all_db_checkpoint_manager):
        """
        Test that we continue with the same pagination mode as was
        already in use
        """

        checkpoint_manager = all_db_checkpoint_manager.for_dataset(
            "case", ["Case"]
        )
        # simulate previous run with legacy pagination mode
        checkpoint_manager.set_checkpoint(
            '2012-04-24T05:13:01', PaginationMode.date_modified, is_final=True
        )

        client = MockCheckpointingClient({
            'case': [({
                'limit': DEFAULT_BATCH_SIZE,
                'order_by': 'server_date_modified',
                'server_date_modified_start': '2012-04-24T05:13:01'
            }, [{
                'id': "doc 1",
                'name': 'n1',
                'server_date_modified': '2012-04-25T05:13:01.000000Z'
            }, {
                'id': "doc 2",
                'name': 'n2',
                'server_date_modified': '2012-04-26T05:13:01.000000Z'
            }])]
        })

        _pull_mock_data(
            writer, all_db_checkpoint_manager, client,
            'tests/013_ConflictingTypes.xlsx'
        )
        self._check_data(writer, [["doc 1"], ["doc 2"]], "Case")
        self._check_checkpoint(
            checkpoint_manager, '2012-04-26T05:13:01', 'doc 2',
            PaginationMode.date_modified.name
        )

    def test_cli_pagination_start_over(
        self, writer, all_db_checkpoint_manager
    ):
        """
        Test that we switch to the new pagination mode when using
        'start_over'
        """
        checkpoint_manager = all_db_checkpoint_manager.for_dataset(
            "case", ["Case"]
        )
        # simulate previous run with legacy pagination mode
        checkpoint_manager.set_checkpoint(
            '2012-04-24T05:13:01', PaginationMode.date_modified, is_final=True
        )

        _pull_mock_data(
            writer,
            all_db_checkpoint_manager,
            get_indexed_on_client(0),
            'tests/013_ConflictingTypes.xlsx',
            start_over=True
        )
        self._check_data(writer, [["doc 1"], ["doc 2"]], "Case")
        self._check_checkpoint(
            checkpoint_manager, '2012-04-24T05:13:01', 'doc 2'
        )

    def test_cli_pagination_since(self, writer, all_db_checkpoint_manager):
        """
        Test that we use to the new pagination mode when using 'since'
        """
        checkpoint_manager = all_db_checkpoint_manager.for_dataset(
            "case", ["Case"]
        )
        # simulate previous run with legacy pagination mode
        checkpoint_manager.set_checkpoint(
            '2012-04-28T05:13:01', PaginationMode.date_modified, is_final=True
        )

        # this will fail if it doesn't use the 'date_indexed' pagination
        # mode due to how the mock client is set up
        _pull_mock_data(
            writer,
            None,
            get_indexed_on_client(1),
            'tests/013_ConflictingTypes.xlsx',
            since='2012-04-24T05:13:01'
        )
        self._check_data(writer, [["doc 3"], ["doc 4"]], "Case")

    def _check_data(self, writer, expected, table_name):
        _check_data(writer, expected, table_name, ['id'])

    def _check_checkpoint(
        self,
        checkpoint_manager,
        since_param,
        doc_id,
        pagination_mode=PaginationMode.date_indexed.name
    ):
        checkpoint = checkpoint_manager.get_last_checkpoint()
        assert checkpoint.pagination_mode == pagination_mode
        assert checkpoint.since_param == since_param
        assert checkpoint.last_doc_id == doc_id



import unittest
from itertools import *

from commcare_export.repeatable_iterator import RepeatableIterator


class TestRepeatableIterator(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        pass

    def test_iteration(self):

        class LazinessException(Exception):
            pass

        def test1():
            for i in range(1, 100):
                yield i

        def test2():
            for i in range(1, 100):
                if i > 10:
                    raise LazinessException('Not lazy enough')
                yield i

        # First make sure that we've properly set up a situation that
        # fails without RepeatableIterator
        iterator = test1()
        assert list(iterator) == list(range(1, 100))
        assert list(iterator) == []

        # Now test that the RepeatableIterator restores functionality
        iterator = RepeatableIterator(test1)
        assert list(iterator) == list(range(1, 100))
        assert list(iterator) == list(range(1, 100))
        assert bool(iterator) is True

        iterator = RepeatableIterator(lambda: (i for i in []))
        assert bool(iterator) is False

        # Ensure that laziness is maintained
        iterator = RepeatableIterator(test2)
        assert list(islice(iterator, 5)) == list(range(1, 6))

        try:
            list(islice(iterator, 15))
            raise Exception('Should have failed')
        except LazinessException:
            pass



class TableSpec:

    def __init__(self, name, headings, rows, data_types=None):
        self.name = name
        self.headings = headings
        self.rows = rows
        self.data_types = data_types or []

    def __eq__(self, other):
        return (
            isinstance(other, TableSpec)
            and other.name == self.name
            and other.headings == self.headings
            and other.data_types == self.data_types
        )

    def toJSON(self):
        return {
            'name': self.name,
            'headings': self.headings,
            'data_types': self.data_types,
        }



import re

from commcare_export import excel_query
from commcare_export.minilinq import Apply, List, Literal, Reference

USERS_TABLE_NAME = 'commcare_users'
LOCATIONS_TABLE_NAME = 'commcare_locations'

excel_query.blacklist(USERS_TABLE_NAME)
excel_query.blacklist(LOCATIONS_TABLE_NAME)


class Column:

    def __init__(self, name, source, map_function=None, *extra_args):
        self.name = Literal(name)
        self.source = source
        self.map_function = map_function
        self.extra_args = extra_args

    @property
    def mapped_source_field(self):
        if not self.map_function:
            return Reference(self.source)
        else:
            return Apply(
                Reference(self.map_function),
                Reference(self.source),
                *self.extra_args
            )


def compile_query(columns, data_source, table_name):
    source = Apply(
        Reference('api_data'),
        Literal(data_source),
        Reference('checkpoint_manager')
    )
    part = excel_query.SheetParts(
        table_name, [c.name for c in columns],
        source,
        List([c.mapped_source_field for c in columns]),
        None
    )
    return excel_query.compile_queries([part], None, False)[0]


# A MiniLinq query for internal CommCare user table. It reads every
# field produced by the /user/ API endpoint and writes the data to a
# table named "commcare_users" in a database.

user_columns = [
    Column('id', 'id'),
    Column('default_phone_number', 'default_phone_number'),
    Column('email', 'email'),
    Column('first_name', 'first_name'),
    Column('groups', 'groups'),
    Column('last_name', 'last_name'),
    Column('phone_numbers', 'phone_numbers'),
    Column('resource_uri', 'resource_uri'),
    Column('commcare_location_id', 'user_data.commcare_location_id'),
    Column('commcare_location_ids', 'user_data.commcare_location_ids'),
    Column(
        'commcare_primary_case_sharing_id',
        'user_data.commcare_primary_case_sharing_id'
    ),
    Column('commcare_project', 'user_data.commcare_project'),
    Column('username', 'username')
]

users_query = compile_query(user_columns, 'user', USERS_TABLE_NAME)

# A MiniLinq query for internal CommCare location table. It reads every
# field produced by the /location/ API endpoint and appends fields to
# hold parent locations using location_type information before writing
# the data to a table named "commcare_locations" in a database.


def get_locations_query(lp):
    location_types = lp.location_types

    # For test stability and clarity, we order location types from
    # deepest to shallowest.
    depth = {}

    def set_depth(lt):
        if lt not in depth:
            parent = location_types[lt]['parent']
            if parent is not None:
                set_depth(parent)
                depth[lt] = depth[parent] + 1
            else:
                depth[lt] = 0

    for lt in location_types:
        set_depth(lt)

    ordered_location_types = sorted(
        location_types.values(), key=lambda lt: -depth[lt['resource_uri']]
    )
    location_codes = [lt['code'] for lt in ordered_location_types]

    # The input names are codes produced by Django's slugify utility
    # method. Replace hyphens with underscores to be easier to use in
    # SQL.
    def sql_column_name(code):
        return re.sub('-', '_', code)

    location_columns = [
        Column('id', 'id'),
        Column('created_at', 'created_at', 'str2date'),
        Column('domain', 'domain'),
        Column('external_id', 'external_id'),
        Column('last_modified', 'last_modified', 'str2date'),
        Column('latitude', 'latitude'),
        Column('location_data', 'location_data', 'json2str'),
        Column('location_id', 'location_id'),
        Column('location_type', 'location_type'),
        Column('longitude', 'longitude'),
        Column('name', 'name'),
        Column('parent', 'parent'),
        Column('resource_uri', 'resource_uri'),
        Column('site_code', 'site_code'),
        Column(
            'location_type_administrative',
            'location_type',
            'get_location_info',
            Literal('administrative')
        ),
        Column(
            'location_type_code',
            'location_type',
            'get_location_info',
            Literal('code')
        ),
        Column(
            'location_type_name',
            'location_type',
            'get_location_info',
            Literal('name')
        ),
        Column(
            'location_type_parent',
            'location_type',
            'get_location_info',
            Literal('parent')
        ),
    ] + [
        Column(
            sql_column_name(code),
            'resource_uri',
            'get_location_ancestor',
            Literal(code)
        ) for code in location_codes
    ]
    return compile_query(location_columns, 'location', LOCATIONS_TABLE_NAME)


# Require specified columns in emitted tables.
class ColumnEnforcer():
    columns_to_require = {
        'form': Column('commcare_userid', '$.metadata.userID'),
        'case': Column('commcare_userid', '$.user_id')
    }

    def __init__(self):
        self._emitted_tables = set([])

    def column_to_require(self, data_source):
        if data_source in ColumnEnforcer.columns_to_require:
            return ColumnEnforcer.columns_to_require[data_source]
        else:
            return None



import glob
import io
import os.path
import re
import sys

import setuptools
from setuptools.command.test import test as TestCommand

VERSION_PATH = 'commcare_export/VERSION'

# Overwrite VERSION if we are actually building for a distribution to pypi
# This code path requires dependencies, etc, to be available
if 'sdist' in sys.argv:
    import commcare_export.version
    with io.open(VERSION_PATH, 'w', encoding='ascii') as fh:
        fh.write(commcare_export.version.git_version())

# This import requires either commcare_export/VERSION or to be in a git clone (as does the package in general)
import commcare_export

version = commcare_export.version.version()

# Crash if the VERSION is not a simple version and it is going to register or upload
if 'register' in sys.argv or 'upload' in sys.argv:
    version = commcare_export.version.stored_version()
    if not version or not re.match('\d+\.\d+\.\d+', version):
        print('Version %s is not an appropriate version for publicizing!' %
              version)
        sys.exit(1)

readme = 'README.md'


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['-vv', '--tb=short']
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(['tests/'] + self.test_args)
        sys.exit(errno)


test_deps = ['pytest', 'psycopg2', 'mock']
base_sql_deps = ["SQLAlchemy", "alembic"]
postgres = ["psycopg2"]
mysql = ["pymysql"]
odbc = ["pyodbc"]

setuptools.setup(
    name="commcare-export",
    version=version,
    description='A command-line tool (and Python library) to extract data from '
    'CommCare HQ into a SQL database or Excel workbook',
    long_description=io.open(readme, encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    author='Dimagi',
    author_email='information@dimagi.com',
    url="https://github.com/dimagi/commcare-export",
    entry_points={
        'console_scripts': [
            'commcare-export = commcare_export.cli:entry_point',
            'commcare-export-utils = commcare_export.utils_cli:entry_point'
        ]
    },
    packages=setuptools.find_packages(exclude=['tests*']),
    data_files=[
        (os.path.join('share', 'commcare-export', 'examples'),
         glob.glob('examples/*.json') + glob.glob('examples/*.xlsx')),
    ],
    include_package_data=True,
    license='MIT',
    python_requires=">=3.6",
    install_requires=[
        'alembic',
        'argparse',
        'backoff>=2.0',
        'jsonpath-ng~=1.6.0',
        'ndg-httpsclient',
        'openpyxl==2.5.12',
        'python-dateutil',
        'pytz',
        'requests',
        'simplejson',
        'sqlalchemy~=1.4',
        'sqlalchemy-migrate'
    ],
    extras_require={
        'test': test_deps,
        'base_sql': base_sql_deps,
        'postgres': base_sql_deps + postgres,
        'mysql': base_sql_deps + mysql,
        'odbc': base_sql_deps + odbc,
        'xlsx': ["openpyxl"],
        'xls': ["xlwt"],
    },
    cmdclass={'test': PyTest},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Database',
        'Topic :: Software Development :: Interpreters',
        'Topic :: System :: Archiving',
        'Topic :: System :: Distributed Computing',
    ]
)



import functools
import hashlib
import inspect
import io

from commcare_export.repeatable_iterator import RepeatableIterator
from jsonpath_ng import jsonpath


def digest_file(path):
    with io.open(path, 'rb') as filehandle:
        digest = hashlib.md5()
        while True:
            # Arbitrary choice of size to be ~filesystem block size friendly
            chunk = filehandle.read(4096)
            if not chunk:
                break
            digest.update(chunk)
        return digest.hexdigest()


def unwrap(arg_name):

    def unwrapper(fn):

        @functools.wraps(fn)
        def _inner(*args):
            callargs = inspect.getcallargs(fn, *args)
            val = callargs[arg_name]
            callargs[arg_name] = unwrap_val(val)
            return fn(**callargs)

        return _inner

    return unwrapper


def unwrap_val(val):
    if isinstance(val, RepeatableIterator):
        val = list(val)

    if isinstance(val, list):
        if len(val) == 1:
            val = val[0]
        else:
            val = [unwrap_val(v) for v in val]

    if isinstance(val, jsonpath.DatumInContext):
        val = val.value

    return val


def default_to_json(obj):
    if hasattr(obj, 'toJSON'):
        return obj.toJSON()
    else:
        return RepeatableIterator.to_jvalue(obj)



import csv
import datetime
import io
import logging
import zipfile
from itertools import zip_longest

import sqlalchemy
from sqlalchemy.exc import NoSuchTableError

from alembic.migration import MigrationContext
from alembic.operations import Operations
from commcare_export.data_types import UnknownDataType, get_sqlalchemy_type
from commcare_export.specs import TableSpec

logger = logging.getLogger(__name__)

MAX_COLUMN_SIZE = 2000


def ensure_text(v, convert_none=False):
    if v is None:
        return '' if convert_none else v

    if isinstance(v, str):
        return v
    elif isinstance(v, bytes):
        return v
    elif isinstance(v, datetime.datetime):
        return v.strftime('%Y-%m-%d %H:%M:%S')
    elif isinstance(v, datetime.date):
        return v.isoformat()
    else:
        return str(v)


def to_jvalue(v):
    if v is None:
        return None

    if isinstance(v, (str, int)):
        return v
    elif isinstance(v, bytes):
        return v
    else:
        return str(v)


class TableWriter(object):
    """
    Interface for export writers: Usable in a "with" statement, and
    while open one can call write_table.

    If the implementing class does not actually need any set up, no-op
    defaults have been provided.
    """
    max_column_length = None
    support_checkpoints = False

    # set to False if writer does not support writing to the same table
    # multiple times
    supports_multi_table_write = True

    required_columns = None

    def __enter__(self):
        return self

    def write_table(self, table: TableSpec) -> None:
        raise NotImplementedError()

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class CsvTableWriter(TableWriter):
    supports_multi_table_write = False

    def __init__(self, file, max_column_size=MAX_COLUMN_SIZE):
        self.file = file
        self.tables = []
        self.archive = None

    def __enter__(self):
        self.archive = zipfile.ZipFile(self.file, 'w', zipfile.ZIP_DEFLATED)
        return self

    def write_table(self, table):
        if self.archive is None:
            raise Exception('Attempt to write to a closed CsvWriter')

        tempfile = io.StringIO()
        writer = csv.writer(tempfile, dialect=csv.excel)
        writer.writerow(table.headings)
        for row in table.rows:
            writer.writerow(row)

        # TODO: make this a polite zip and put everything in a subfolder
        #       with the same basename as the zipfile
        self.archive.writestr(
            '%s.csv' % self.zip_safe_name(table.name),
            tempfile.getvalue().encode('utf-8')
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.archive.close()

    def zip_safe_name(self, name):
        return name[:31]


class Excel2007TableWriter(TableWriter):
    max_table_name_size = 31

    def __init__(self, file):
        try:
            import openpyxl
        except ImportError:
            raise Exception(
                "It doesn't look like this machine is configured for "
                "Excel export. To export to Excel you have to run the "
                "command:  pip install openpyxl"
            )

        self.file = file
        self.book = openpyxl.workbook.Workbook(write_only=True)
        self.sheets = {}

    def __enter__(self):
        return self

    def write_table(self, table):
        sheet = self.get_sheet(table)
        for row in table.rows:
            sheet.append([ensure_text(v) for v in row])

    def get_sheet(self, table):
        name = table.name
        if name not in self.sheets:
            sheet = self.book.create_sheet()
            sheet.title = name[:self.max_table_name_size]
            sheet.append([ensure_text(v) for v in table.headings])
            self.sheets[name] = sheet

        return self.sheets[name]

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.book.save(self.file)


class Excel2003TableWriter(TableWriter):
    max_table_name_size = 31

    def __init__(self, file):
        try:
            import xlwt
        except ImportError:
            raise Exception(
                "It doesn't look like this machine is configured for "
                "excel export. To export to excel you have to run the "
                "command:  pip install xlwt"
            )

        self.file = file
        self.book = xlwt.Workbook()
        self.sheets = {}

    def __enter__(self):
        return self

    def write_table(self, table):
        sheet, current_row = self.get_sheet(table)
        for row in table.rows:
            for colnum, val in enumerate(row):
                sheet.write(current_row, colnum, ensure_text(val))
            current_row += 1

        self.sheets[table.name] = (sheet, current_row)

    def get_sheet(self, table):
        name = table.name
        if name not in self.sheets:
            sheet = self.book.add_sheet(name[:self.max_table_name_size])
            sheet.title = name[:self.max_table_name_size]

            for colnum, val in enumerate(table.headings):
                sheet.write(0, colnum, ensure_text(val))

            self.sheets[name] = (sheet, 1)  # start from row 1

        return self.sheets[name]

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.book.save(self.file)


class JValueTableWriter(TableWriter):
    """
    Write tables to JSON-friendly in-memory values
    """

    def __init__(self):
        self.tables = {}

    def write_table(self, table):
        if table.name not in self.tables:
            self.tables[table.name] = TableSpec(
                name=table.name,
                headings=list(table.headings),
                rows=[],
            )
        else:
            assert self.tables[table.name].headings == list(table.headings)

        self.tables[table.name].rows = list(
            self.tables[table.name].rows
        ) + [[to_jvalue(v) for v in row] for row in table.rows]


class StreamingMarkdownTableWriter(TableWriter):
    """
    Writes markdown to an output stream, where each table just comes one
    after the other
    """
    supports_multi_table_write = False

    def __init__(self, output_stream, compute_widths=False):
        self.output_stream = output_stream
        self.compute_widths = compute_widths

    def write_table(self, table):
        col_widths = None
        if self.compute_widths:
            col_widths = self._get_column_widths(table)
            row_template = ' | '.join([
                '{{:<{}}}'.format(width) for width in col_widths
            ])
        else:
            row_template = ' | '.join(['{}'] * len(table.headings))

        if table.name:
            self.output_stream.write('\n# %s \n\n' % table.name)

        self.output_stream.write(
            '| %s |\n' % row_template.format(*table.headings)
        )
        if col_widths:
            self.output_stream.write(
                '| %s |\n'
                % row_template.format(*['-' * width for width in col_widths])
            )

        for row in table.rows:
            text_row = (ensure_text(val, convert_none=True) for val in row)
            self.output_stream.write(
                '| %s |\n' % row_template.format(*text_row)
            )

    def _get_column_widths(self, table):
        all_rows = [table.headings] + table.rows
        columns = list(map(list, zip(*all_rows)))
        col_widths = map(len, [max(col, key=len) for col in columns])
        return list(col_widths)


class SqlMixin(object):
    """
    Write tables to a database specified by URL
    (TODO) with "upsert" based on primary key.
    """

    MIN_VARCHAR_LEN = 32
    # Arbitrary point at which we switch to TEXT; for Postgres
    # VARCHAR == TEXT anyhow
    MAX_VARCHAR_LEN = 255

    def __init__(self, db_url, poolclass=None, engine=None):
        self.db_url = db_url
        self.collation = 'utf8mb4_unicode_ci' if 'mysql' in db_url else None
        self.engine = engine or sqlalchemy.create_engine(
            db_url, poolclass=poolclass
        )
        self._metadata = None

    def __enter__(self):
        self.connection = self.engine.connect()
        return self  # TODO: fork the writer so this can be called many times

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connection.close()

    @property
    def is_postgres(self):
        return 'postgres' in self.db_url

    @property
    def is_oracle(self):
        return 'oracle' in self.db_url

    @property
    def is_mysql(self):
        return 'mysql' in self.db_url

    @property
    def is_mssql(self):
        return 'mssql' in self.db_url

    @property
    def max_column_length(self):
        if self.is_postgres:
            # https://www.postgresql.org/docs/current/static/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS
            return 63
        if self.is_mysql:
            # https://dev.mysql.com/doc/refman/8.0/en/identifiers.html
            return 64
        if self.is_mssql:
            # https://docs.microsoft.com/en-us/sql/relational-databases/databases/database-identifiers?view=sql-server-2017
            return 128
        if self.is_oracle:
            return 128
        raise Exception("Unknown database dialect: {}".format(self.db_url))

    @property
    def metadata(self):
        if (
            self._metadata is None
            or self._metadata.bind.closed
            or self._metadata.bind.invalidated
        ):
            if self.connection.closed:
                raise Exception('Tried to bind to a closed connection')
            if self.connection.invalidated:
                raise Exception('Tried to bind to an invalidated connection')
            self._metadata = sqlalchemy.MetaData(bind=self.connection)
        return self._metadata

    def get_table(self, table_name):
        try:
            return sqlalchemy.Table(
                table_name,
                self.metadata,
                autoload_with=self.connection,
            )
        except NoSuchTableError:
            return None

    def get_id_column(self):
        return sqlalchemy.Column(
            'id', sqlalchemy.Unicode(self.MAX_VARCHAR_LEN), primary_key=True
        )


class SqlTableWriter(SqlMixin, TableWriter):
    """
    Write tables to a database specified by URL
    (TODO) with "upsert" based on primary key.
    """
    support_checkpoints = True
    required_columns = ['id']

    def __init__(self, db_url, strict_types=False, poolclass=None):
        super(SqlTableWriter, self).__init__(db_url, poolclass=poolclass)
        self.strict_types = strict_types

    def get_data_type(self, explicit_type, val):
        if explicit_type:
            return self.get_explicit_type(explicit_type)
        return self.best_type_for(val)

    def get_explicit_type(self, data_type):
        try:
            return get_sqlalchemy_type(data_type)
        except UnknownDataType:
            if data_type:
                logger.warning(
                    "Found unknown data type '{data_type}'".format(
                        data_type=data_type,
                    )
                )
            return self.best_type_for('')  # todo: more explicit fallback

    def best_type_for(self, val):
        if isinstance(val, bool):
            return sqlalchemy.Boolean()
        elif isinstance(val, datetime.datetime):
            if self.is_mssql:
                return sqlalchemy.dialects.mssql.DATETIME2()
            else:
                return sqlalchemy.DateTime()
        elif isinstance(val, datetime.date):
            return sqlalchemy.Date()

        if isinstance(val, int):
            return sqlalchemy.Integer()
        elif isinstance(val, str):
            if self.is_postgres:
                # PostgreSQL is the best; you can use TEXT everywhere
                # and it works like a charm.
                return sqlalchemy.UnicodeText(collation=self.collation)
            elif self.is_mysql:
                # MySQL cannot build an index on TEXT due to the lack of
                # a field length, so we try to use VARCHAR when
                # possible.
                if len(val) < self.MAX_VARCHAR_LEN:
                    return sqlalchemy.Unicode(
                        max(len(val), self.MIN_VARCHAR_LEN),
                        collation=self.collation
                    )
                else:
                    return sqlalchemy.UnicodeText(collation=self.collation)
            elif self.is_mssql:
                # MSSQL (pre 2016) doesn't allow indices on columns
                # longer than 900 bytes
                # https://docs.microsoft.com/en-us/sql/t-sql/statements/create-index-transact-sql
                # If any of our data is bigger than this, then set the
                # column to NVARCHAR(max) `length` here is the size in
                # bytes
                # https://docs.sqlalchemy.org/en/13/core/type_basics.html#sqlalchemy.types.String.params.length
                length_in_bytes = len(val.encode('utf-8'))
                column_length_in_bytes = None if length_in_bytes > 900 else 900
                return sqlalchemy.NVARCHAR(
                    length=column_length_in_bytes, collation=self.collation
                )
            elif self.is_oracle:
                return sqlalchemy.Unicode(4000, collation=self.collation)
            else:
                raise Exception(f"Unknown database dialect: {self.db_url}")
        else:
            # We do not have a name for "bottom" in SQL aka the type
            # whose least upper bound with any other type is the other
            # type.
            return sqlalchemy.UnicodeText(collation=self.collation)

    def compatible(self, source_type, dest_type):
        """
        Checks _coercion_ compatibility.
        """
        if isinstance(source_type, sqlalchemy.String):
            if not isinstance(dest_type, sqlalchemy.String):
                return False
            elif source_type.length is None:
                # The length being None means that we are looking at
                # indefinite strings aka TEXT. This tool will never
                # create strings with bounds, but if a target DB has one
                # then we cannot insert to it. We will request that
                # whoever uses this tool convert to TEXT type.
                return dest_type.length is None
            else:
                return dest_type.length is None or (
                    dest_type.length >= source_type.length
                )

        compatibility = {
            sqlalchemy.String: (sqlalchemy.Text,),
            sqlalchemy.Integer: (sqlalchemy.String, sqlalchemy.Text),
            sqlalchemy.Boolean:
                (sqlalchemy.String, sqlalchemy.Text, sqlalchemy.Integer),
            sqlalchemy.DateTime:
                (sqlalchemy.String, sqlalchemy.Text, sqlalchemy.Date),
            sqlalchemy.Date: (sqlalchemy.String, sqlalchemy.Text),
        }

        # add dialect specific types
        try:
            compatibility[sqlalchemy.JSON
                         ] = (sqlalchemy.dialects.postgresql.json.JSON,)
        except AttributeError:
            pass
        try:
            compatibility[sqlalchemy.Boolean
                         ] += (sqlalchemy.dialects.mssql.base.BIT,)
        except AttributeError:
            pass

        for _type, types in compatibility.items():
            if isinstance(source_type, _type):
                return isinstance(dest_type, (_type,) + types)

    def strict_types_compatibility_check(self, source_type, dest_type, val):
        if isinstance(source_type, sqlalchemy.String):
            if not isinstance(dest_type, sqlalchemy.String):
                return  # Can't do anything
            elif dest_type.length is None:
                return  # already a TEXT column
            elif isinstance(val, str) and dest_type.length >= len(val):
                return  # no need to upgrade to TEXT column
            elif source_type.length is None:
                return sqlalchemy.UnicodeText(collation=self.collation)
            elif dest_type.length < source_type.length:
                return source_type

    def least_upper_bound(self, source_type, dest_type):
        """
        Returns the _coercion_ least upper bound.

        Mostly just promotes everything to string if it is not already.
        In fact, since this is only called when they are incompatible,
        it promotes to string right away.
        """

        # FIXME: Don't be so silly
        return sqlalchemy.UnicodeText(collation=self.collation)

    def make_table_compatible(self, table, row_dict, data_type_dict):
        ctx = MigrationContext.configure(self.connection)
        op = Operations(ctx)
        columns = {c.name: c for c in table.columns}

        for column, val in row_dict.items():
            if val is None:
                continue

            val_type = self.get_data_type(data_type_dict[column], val)
            if column not in columns:
                logger.warning(
                    f"Adding column '{table.name}.{column} {val_type}'"
                )
                op.add_column(
                    table.name,
                    sqlalchemy.Column(column, val_type, nullable=True)
                )
                self.metadata.clear()
                table = self.get_table(table.name)
            elif not columns[column].primary_key:
                col_type = columns[column].type
                new_col_type = None
                if self.strict_types:
                    # don't bother checking compatibility since we're
                    # not going to change anything
                    new_col_type = self.strict_types_compatibility_check(
                        val_type, col_type, val
                    )
                elif not self.compatible(val_type, col_type):
                    new_col_type = self.least_upper_bound(val_type, col_type)

                if new_col_type:
                    logger.warning(
                        f'Altering column {columns[column]} from {col_type} '
                        f'to {new_col_type} for value: "{type(val)}:{val}"',
                    )
                    op.alter_column(table.name, column, type_=new_col_type)
                    self.metadata.clear()
                    table = self.get_table(table.name)
        return table

    def create_table(self, table_name, row_dict, data_type_dict):
        ctx = MigrationContext.configure(self.connection)
        op = Operations(ctx)
        if self.strict_types:
            create_sql = sqlalchemy.schema.CreateTable(
                sqlalchemy.Table(
                    table_name,
                    sqlalchemy.MetaData(),
                    *self._get_columns_for_data(row_dict, data_type_dict)
                )
            ).compile(self.connection.engine)
            logger.warning(
                f"Table '{table_name}' does not exist. Creating table "
                f"with:\n{create_sql}"
            )
            empty_cols = [
                name for (name, val) in row_dict.items()
                if val is None and name not in data_type_dict
            ]
            if empty_cols:
                logger.warning(
                    "This schema does not include the following columns "
                    "since we are unable to determine the column type at "
                    f"this stage: {empty_cols}"
                )
        op.create_table(
            table_name,
            *self._get_columns_for_data(row_dict, data_type_dict)
        )
        self.metadata.clear()
        return self.get_table(table_name)

    def upsert(self, table, row_dict):
        # For atomicity "insert, catch, update" is slightly better than
        # "select, insert or update". The latter may crash, while the
        # former may overwrite data (which should be fine if whatever is
        # racing against this is importing from the same source... if
        # not you are busted anyhow

        # strip out values that are None since the column may not exist
        # yet
        row_dict = {
            col: val for col, val in row_dict.items() if val is not None
        }
        try:
            insert = table.insert().values(**row_dict)
            self.connection.execute(insert)
        except sqlalchemy.exc.IntegrityError:
            update = (table.update()
                      .where(table.c.id == row_dict['id'])
                      .values(**row_dict))
            self.connection.execute(update)

    def write_table(self, table_spec: TableSpec) -> None:
        table_name = table_spec.name
        headings = table_spec.headings
        data_type_dict = dict(zip_longest(headings, table_spec.data_types))
        for i, row in enumerate(table_spec.rows):
            row_dict = dict(zip(headings, row))
            if i == 0:
                table = self.get_table(table_name)
                if table is None:
                    table = self.create_table(
                        table_name,
                        row_dict,
                        data_type_dict,
                    )
            # Checks the data type for every cell in every row. Maybe we
            # can use a future version of the data dictionary to avoid
            # this?
            table = self.make_table_compatible(table, row_dict, data_type_dict)
            self.upsert(table, row_dict)

    def _get_columns_for_data(self, row_dict, data_type_dict):
        return [self.get_id_column()] + [
            sqlalchemy.Column(
                column_name,
                self.get_data_type(data_type_dict[column_name], val),
                nullable=True
            )
            for (column_name, val) in row_dict.items()
            if ((val is not None or data_type_dict[column_name])
                and column_name != 'id')
        ]



import argparse
import getpass
import io
import json
import logging
import os.path
import sys

import dateutil.parser
import requests
import sqlalchemy

from commcare_export import builtin_queries, excel_query, writers
from commcare_export.checkpoint import CheckpointManagerProvider
from commcare_export.commcare_hq_client import (
    LATEST_KNOWN_VERSION,
    CommCareHqClient,
    ResourceRepeatException,
)
from commcare_export.commcare_minilinq import CommCareHqEnv
from commcare_export.env import BuiltInEnv, EmitterEnv, JsonPathEnv
from commcare_export.exceptions import (
    DataExportException,
    MissingQueryFileException,
)
from commcare_export.location_info_provider import LocationInfoProvider
from commcare_export.minilinq import List, MiniLinq
from commcare_export.misc import default_to_json
from commcare_export.repeatable_iterator import RepeatableIterator
from commcare_export.utils import get_checkpoint_manager
from commcare_export.version import __version__

EXIT_STATUS_ERROR = 1

logger = logging.getLogger(__name__)

commcare_hq_aliases = {
    'local': 'http://localhost:8000',
    'prod': 'https://www.commcarehq.org'
}


class Argument(object):

    def __init__(self, name, *args, **kwargs):
        self.name = name.replace('-', '_')
        self._args = ['--{}'.format(name)] + list(args)
        self._kwargs = kwargs

    @property
    def default(self):
        return self._kwargs.get('default')

    def add_to_parser(self, parser, **additional_kwargs):
        additional_kwargs.update(self._kwargs)
        parser.add_argument(*self._args, **additional_kwargs)


CLI_ARGS = [
    Argument(
        'version',
        default=False,
        action='store_true',
        help='Print the current version of the commcare-export tool.'
    ),
    Argument('query', required=False, help='JSON or Excel query file'),
    Argument('dump-query', default=False, action='store_true'),
    Argument(
        'commcare-hq',
        default='prod',
        help='Base url for the CommCare HQ instance e.g. '
        'https://www.commcarehq.org'
    ),
    Argument('api-version', default=LATEST_KNOWN_VERSION),
    Argument('project'),
    Argument('username'),
    Argument(
        'password',
        help='Enter password, or if using apikey auth-mode, enter the api key.'
    ),
    Argument(
        'auth-mode',
        default='password',
        choices=['password', 'apikey'],
        help='Use "digest" auth, or "apikey" auth (for two factor enabled '
        'domains).'
    ),
    Argument(
        'since',
        help='Export all data after this date. Format YYYY-MM-DD or '
        'YYYY-MM-DDTHH:mm:SS'
    ),
    Argument(
        'until',
        help='Export all data up until this date. Format YYYY-MM-DD or '
        'YYYY-MM-DDTHH:mm:SS'
    ),
    Argument(
        'start-over',
        default=False,
        action='store_true',
        help='When saving to a SQL database; the default is to pick up '
        'since the last success. This disables that.'
    ),
    Argument('profile'),
    Argument('verbose', default=False, action='store_true'),
    Argument(
        'output-format',
        default='json',
        choices=['json', 'csv', 'xls', 'xlsx', 'sql', 'markdown'],
        help='Output format'
    ),
    Argument(
        'output',
        metavar='PATH',
        default='reports.zip',
        help='Path to output; defaults to `reports.zip`.'
    ),
    Argument(
        'strict-types',
        default=False,
        action='store_true',
        help="When saving to a SQL database don't allow changing column types "
        "once they are created."
    ),
    Argument(
        'missing-value',
        default=None,
        help="Value to use when a field is missing from the form / case."
    ),
    Argument(
        'batch-size',
        default=200,
        help="Number of records to process per batch."
    ),
    Argument(
        'checkpoint-key',
        help="Use this key for all checkpoints instead of the query file MD5 "
        "hash in order to prevent table rebuilds after a query file has "
        "been edited."
    ),
    Argument(
        'users',
        default=False,
        action='store_true',
        help="Export a table containing data about this project's mobile "
        "workers"
    ),
    Argument(
        'locations',
        default=False,
        action='store_true',
        help="Export a table containing data about this project's locations"
    ),
    Argument(
        'with-organization',
        default=False,
        action='store_true',
        help="Export tables containing mobile worker data and location data "
        "and add a commcare_userid field to any exported form or case"
    ),
    Argument(
        'export-root-if-no-subdocument',
        default=False,
        action='store_true',
        help="Use this when you are exporting a nested document e.g. "
        "form.form..case, messaging-event.messages.[*] And you want to "
        "have a record exported even if the nested document does not "
        "exist or is empty.",
    )
]


def main(argv):
    parser = argparse.ArgumentParser(
        'commcare-export', 'Output a customized export of CommCareHQ data.'
    )
    for arg in CLI_ARGS:
        arg.add_to_parser(parser)

    args = parser.parse_args(argv)

    if args.verbose:
        logging.basicConfig(
            level=logging.DEBUG,
            format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s'
        )
    else:
        logging.basicConfig(
            level=logging.WARN,
            format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s'
        )

    logging.getLogger('alembic').setLevel(logging.WARN)
    logging.getLogger('backoff').setLevel(logging.FATAL)
    logging.getLogger('urllib3').setLevel(logging.WARN)

    if args.version:
        print('commcare-export version {}'.format(__version__))
        exit(0)

    if not args.project:
        print(
            'commcare-export: error: argument --project is required',
            file=sys.stderr
        )
        exit(1)

    if args.profile:
        # hotshot is gone in Python 3
        import hotshot
        import hotshot.stats
        profile = hotshot.Profile(args.profile)
        profile.start()

    try:
        exit(main_with_args(args))
    finally:
        if args.profile:
            profile.close()
            stats = hotshot.stats.load(args.profile)
            stats.strip_dirs()
            stats.sort_stats('cumulative', 'calls')
            stats.print_stats(100)


def _get_query(args, writer, column_enforcer=None):
    return _get_query_from_file(
        args.query,
        args.missing_value,
        writer.supports_multi_table_write,
        writer.max_column_length,
        writer.required_columns,
        column_enforcer,
        args.export_root_if_no_subdocument
    )


def _get_query_from_file(
    query_arg,
    missing_value,
    combine_emits,
    max_column_length,
    required_columns,
    column_enforcer,
    value_or_root
):
    if os.path.exists(query_arg):
        if os.path.splitext(query_arg)[1] in ['.xls', '.xlsx']:
            import openpyxl
            workbook = openpyxl.load_workbook(query_arg)
            return excel_query.get_queries_from_excel(
                workbook,
                missing_value,
                combine_emits,
                max_column_length,
                required_columns,
                column_enforcer,
                value_or_root
            )
        else:
            with io.open(query_arg, encoding='utf-8') as fh:
                return MiniLinq.from_jvalue(json.loads(fh.read()))


def get_queries(args, writer, lp, column_enforcer=None):
    query_list = []
    if args.query is not None:
        query = _get_query(args, writer, column_enforcer=column_enforcer)

        if not query:
            raise MissingQueryFileException(args.query)
        query_list.append(query)

    if args.users or args.with_organization:
        # Add user data to query
        query_list.append(builtin_queries.users_query)

    if args.locations or args.with_organization:
        # Add location data to query
        query_list.append(builtin_queries.get_locations_query(lp))

    return List(query_list) if len(query_list) > 1 else query_list[0]


def _get_writer(output_format, output, strict_types):
    if output_format == 'xlsx':
        return writers.Excel2007TableWriter(output)
    elif output_format == 'xls':
        return writers.Excel2003TableWriter(output)
    elif output_format == 'csv':
        if not output.endswith(".zip"):
            print(
                "WARNING: csv output is a zip file, but "
                "will be written to %s" % output
            )
            print(
                "Consider appending .zip to the file name to avoid confusion."
            )
        return writers.CsvTableWriter(output)
    elif output_format == 'json':
        return writers.JValueTableWriter()
    elif output_format == 'markdown':
        return writers.StreamingMarkdownTableWriter(sys.stdout)
    elif output_format == 'sql':
        # Output should be a connection URL. Writer had bizarre issues
        # so we use a full connection instead of passing in a URL or
        # engine.
        if output.startswith('mysql'):
            charset_split = output.split('charset=')
            if len(charset_split) > 1 and charset_split[1] != 'utf8mb4':
                raise Exception(
                    f"The charset '{charset_split[1]}' might cause problems with the export. "
                    f"It is recommended that you use 'utf8mb4' instead."
                )

        return writers.SqlTableWriter(output, strict_types)
    else:
        raise Exception("Unknown output format: {}".format(output_format))


def get_date_params(args):
    since = dateutil.parser.parse(args.since) if args.since else None
    until = dateutil.parser.parse(args.until) if args.until else None
    return since, until


def _get_api_client(args, commcarehq_base_url):
    return CommCareHqClient(
        url=commcarehq_base_url,
        project=args.project,
        username=args.username,
        password=args.password,
        auth_mode=args.auth_mode,
        version=args.api_version
    )


def _get_checkpoint_manager(args):
    if not args.users and not args.locations and not os.path.exists(
        args.query
    ):
        logger.warning(
            "Checkpointing disabled for non builtin, "
            "non file-based query"
        )
    elif args.since or args.until:
        logger.warning(
            "Checkpointing disabled when using '--since' or '--until'"
        )
    else:
        checkpoint_manager = get_checkpoint_manager(args)
        checkpoint_manager.create_checkpoint_table()
        return checkpoint_manager


def force_lazy_result(lazy_result):
    if lazy_result is not None:
        if isinstance(lazy_result, RepeatableIterator):
            list(lazy_result) if lazy_result else lazy_result
        else:
            for nested_result in lazy_result:
                force_lazy_result(nested_result)


def evaluate_query(env, query):
    with env:
        try:
            lazy_result = query.eval(env)
            force_lazy_result(lazy_result)
            return 0
        except requests.exceptions.RequestException as e:
            if e.response and e.response.status_code == 401:
                print(
                    "\nAuthentication failed. Please check your credentials.",
                    file=sys.stderr
                )
                return EXIT_STATUS_ERROR
            else:
                raise
        except ResourceRepeatException as e:
            print('Stopping because the export is stuck')
            print(e.message)
            print('Try increasing --batch-size to overcome the error')
            return EXIT_STATUS_ERROR
        except (
            sqlalchemy.exc.DataError,
            sqlalchemy.exc.InternalError,
            sqlalchemy.exc.ProgrammingError
        ) as e:
            print('Stopping because of database error:\n', e)
            return EXIT_STATUS_ERROR
        except KeyboardInterrupt:
            print('\nExport aborted', file=sys.stderr)
            return EXIT_STATUS_ERROR


def main_with_args(args):
    logger.info("CommCare Export Version {}".format(__version__))
    writer = _get_writer(args.output_format, args.output, args.strict_types)

    if args.query is None and args.users is False and args.locations is False:
        print(
            'At least one the following arguments is required: '
            '--query, --users, --locations',
            file=sys.stderr
        )
        return EXIT_STATUS_ERROR

    if not args.username:
        args.username = input('Please provide a username: ')

    if not args.password:
        # Windows getpass does not accept unicode
        args.password = getpass.getpass()

    column_enforcer = None
    if args.with_organization:
        column_enforcer = builtin_queries.ColumnEnforcer()

    commcarehq_base_url = commcare_hq_aliases.get(
        args.commcare_hq, args.commcare_hq
    )
    api_client = _get_api_client(args, commcarehq_base_url)
    lp = LocationInfoProvider(api_client, page_size=args.batch_size)
    try:
        query = get_queries(args, writer, lp, column_enforcer)
    except DataExportException as e:
        print(e.message, file=sys.stderr)
        return EXIT_STATUS_ERROR

    if args.dump_query:
        print(json.dumps(query.to_jvalue(), indent=4))
        return

    checkpoint_manager = None
    if writer.support_checkpoints:
        checkpoint_manager = _get_checkpoint_manager(args)

    since, until = get_date_params(args)
    if args.start_over:
        if checkpoint_manager:
            logger.warning(
                'Ignoring all checkpoints and re-fetching all data from '
                'CommCare.'
            )
    elif since:
        logger.debug('Starting from %s', args.since)

    cm = CheckpointManagerProvider(checkpoint_manager, since, args.start_over)
    static_env = {
        'commcarehq_base_url': commcarehq_base_url,
        'get_checkpoint_manager': cm.get_checkpoint_manager,
        'get_location_info': lp.get_location_info,
        'get_location_ancestor': lp.get_location_ancestor
    }
    env = (
        BuiltInEnv(static_env)
        | CommCareHqEnv(api_client, until=until, page_size=args.batch_size)
        | JsonPathEnv({})
        | EmitterEnv(writer)
    )

    exit_status = evaluate_query(env, query)

    if args.output_format == 'json':
        print(
            json.dumps(
                list(writer.tables.values()),
                indent=4,
                default=default_to_json
            )
        )

    return exit_status


def entry_point():
    main(sys.argv[1:])


if __name__ == '__main__':
    entry_point()



"""add key column

Revision ID: d82e1d06a82c
Revises: 53f1aad98e33
Create Date: 2018-08-15 12:08:51.720796

"""
from alembic import op
import sqlalchemy as sa


revision = 'd82e1d06a82c'
down_revision = '53f1aad98e33'
branch_labels = None
depends_on = None


def upgrade():
    url = op.get_bind().engine.url
    collation = 'utf8_bin' if 'mysql' in url.drivername else None
    op.add_column(
        'commcare_export_runs',
        sa.Column('key', sa.Unicode(255, collation=collation))
    )

def downgrade():
    op.drop_column('commcare_export_runs', 'key')



import unittest
from itertools import *

from commcare_export.checkpoint import CheckpointManagerWithDetails
from commcare_export.commcare_hq_client import MockCommCareHqClient
from commcare_export.commcare_minilinq import *
from commcare_export.env import *
from commcare_export.minilinq import *
from jsonpath_ng import jsonpath


class TestCommCareMiniLinq(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        pass

    def check_case(self, val, result):
        if isinstance(result, list):
            assert [
                datum.value
                if isinstance(datum, jsonpath.DatumInContext) else datum
                for datum in val
            ] == result

    def test_eval_indexed_on(self):
        self._test_eval(PaginationMode.date_indexed)

    def test_eval_modified_on(self):
        self._test_eval(PaginationMode.date_modified)

    def _test_eval(self, pagination_mode):
        form_order_by = get_paginator(
            'form', pagination_mode=pagination_mode
        ).since_field
        case_order_by = get_paginator(
            'case', pagination_mode=pagination_mode
        ).since_field

        def die(msg):
            raise Exception(msg)

        client = MockCommCareHqClient({
            'form': [
                (
                    {
                        'limit': 1000,
                        'filter': 'test1',
                        'order_by': form_order_by
                    },
                    [1, 2, 3],
                ),
                ({
                    'limit': 1000,
                    'filter': 'test2',
                    'order_by': form_order_by
                }, [{
                    'x': [{
                        'y': 1
                    }, {
                        'y': 2
                    }]
                }, {
                    'x': [{
                        'y': 3
                    }, {
                        'z': 4
                    }]
                }, {
                    'x': [{
                        'y': 5
                    }]
                }]),
                ({
                    'limit': 1000,
                    'filter': 'laziness-test',
                    'order_by': form_order_by
                },
                 (i if i < 5 else die('Not lazy enough') for i in range(12))),
                ({
                    'limit': 1000,
                    'cases__full': 'true',
                    'order_by': form_order_by
                }, [1, 2, 3, 4, 5]),
            ],
            'case': [({
                'limit': 1000,
                'type': 'foo',
                'order_by': case_order_by
            }, [
                {
                    'x': 1
                },
                {
                    'x': 2
                },
                {
                    'x': 3
                },
            ])],
            'user': [({
                'limit': 1000
            }, [
                {
                    'x': 1
                },
                {
                    'x': 2
                },
                {
                    'x': 3
                },
            ])]
        })

        env = BuiltInEnv() | CommCareHqEnv(client) | JsonPathEnv(
            {}
        )  # {'form': api_client.iterate('form')})

        checkpoint_manager = CheckpointManagerWithDetails(
            None, None, pagination_mode
        )
        assert list(
            Apply(
                Reference('api_data'), Literal('form'),
                Literal(checkpoint_manager), Literal({"filter": 'test1'})
            ).eval(env)
        ) == [1, 2, 3]

        # just check that we can still apply some deeper xpath by
        # mapping; first ensure the basics work
        assert list(
            Apply(
                Reference('api_data'), Literal('form'),
                Literal(checkpoint_manager), Literal({"filter": 'test2'})
            ).eval(env)
        ) == [{
            'x': [{
                'y': 1
            }, {
                'y': 2
            }]
        }, {
            'x': [{
                'y': 3
            }, {
                'z': 4
            }]
        }, {
            'x': [{
                'y': 5
            }]
        }]

        self.check_case(
            FlatMap(
                source=Apply(
                    Reference('api_data'), Literal('form'),
                    Literal(checkpoint_manager), Literal({"filter": 'test2'})
                ),
                body=Reference('x[*].y')
            ).eval(env), [1, 2, 3, 5]
        )

        self.check_case(
            islice(
                Apply(
                    Reference('api_data'), Literal('form'),
                    Literal(checkpoint_manager),
                    Literal({"filter": "laziness-test"})
                ).eval(env), 5
            ), [0, 1, 2, 3, 4]
        )

        self.check_case(
            Apply(
                Reference('api_data'), Literal('form'),
                Literal(checkpoint_manager), Literal(None), Literal(['cases'])
            ).eval(env), [1, 2, 3, 4, 5]
        )

        self.check_case(
            FlatMap(
                source=Apply(
                    Reference('api_data'), Literal('case'),
                    Literal(checkpoint_manager), Literal({'type': 'foo'})
                ),
                body=Reference('x')
            ).eval(env), [1, 2, 3]
        )

        self.check_case(
            FlatMap(
                source=Apply(
                    Reference('api_data'), Literal('user'),
                    Literal(checkpoint_manager), Literal(None)
                ),
                body=Reference('x')
            ).eval(env), [1, 2, 3]
        )



import csv
import datetime
import io
import tempfile
import zipfile

import openpyxl
import sqlalchemy

import pytest
from commcare_export.specs import TableSpec
from commcare_export.writers import (
    CsvTableWriter,
    Excel2007TableWriter,
    JValueTableWriter,
    SqlTableWriter,
)


@pytest.fixture()
def writer(db_params):
    return SqlTableWriter(db_params['url'], poolclass=sqlalchemy.pool.NullPool)


@pytest.fixture()
def strict_writer(db_params):
    return SqlTableWriter(
        db_params['url'],
        poolclass=sqlalchemy.pool.NullPool,
        strict_types=True
    )


TYPE_MAP = {
    'mysql': {
        bool: lambda x: int(x)
    },
}


class TestWriters(object):

    def test_JValueTableWriter(self):
        writer = JValueTableWriter()
        writer.write_table(
            TableSpec(
                **{
                    'name':
                        'foo',
                    'headings': ['a', 'bjørn', 'c', 'd'],
                    'rows': [
                        [1, '2', 3, datetime.date(2015, 1, 1)],
                        [4, '日本', 6, datetime.date(2015, 1, 2)],
                    ]
                }
            )
        )

        writer.write_table(
            TableSpec(
                **{
                    'name': 'foo',
                    'headings': ['a', 'bjørn', 'c', 'd'],
                    'rows': [[5, 'bob', 9,
                              datetime.date(2018, 1, 2)],]
                }
            )
        )

        assert writer.tables == {
            'foo':
                TableSpec(
                    **{
                        'name':
                            'foo',
                        'headings': ['a', 'bjørn', 'c', 'd'],
                        'rows': [
                            [1, '2', 3, '2015-01-01'],
                            [4, '日本', 6, '2015-01-02'],
                            [5, 'bob', 9, '2018-01-02'],
                        ],
                    }
                )
        }

    def test_Excel2007TableWriter(self):
        with tempfile.NamedTemporaryFile(suffix='.xlsx') as file:
            with Excel2007TableWriter(file=file) as writer:
                writer.write_table(
                    TableSpec(
                        **{
                            'name': 'foo',
                            'headings': ['a', 'bjørn', 'c'],
                            'rows': [
                                [1, '2', 3],
                                [4, '日本', 6],
                            ]
                        }
                    )
                )

            self._check_Excel2007TableWriter_output(file.name)

    def test_Excel2007TableWriter_write_mutli(self):
        with tempfile.NamedTemporaryFile(suffix='.xlsx') as file:
            with Excel2007TableWriter(file=file) as writer:
                writer.write_table(
                    TableSpec(
                        **{
                            'name': 'foo',
                            'headings': ['a', 'bjørn', 'c'],
                            'rows': [[1, '2', 3],]
                        }
                    )
                )

                writer.write_table(
                    TableSpec(
                        **{
                            'name': 'foo',
                            'headings': ['a', 'bjørn', 'c'],
                            'rows': [[4, '日本', 6],]
                        }
                    )
                )
            self._check_Excel2007TableWriter_output(file.name)

    def _check_Excel2007TableWriter_output(self, filename):
        output_wb = openpyxl.load_workbook(filename)

        assert output_wb.sheetnames == ['foo']
        foo_sheet = output_wb['foo']
        assert [
            [cell.value for cell in row] for row in foo_sheet['A1:C3']
        ] == [
            ['a', 'bjørn', 'c'],
            ['1', '2', '3'
            ],  # Note how pyxl does some best-effort parsing to *whatever* type
            ['4', '日本', '6'],
        ]

    def test_CsvTableWriter(self):
        with tempfile.NamedTemporaryFile() as file:
            with CsvTableWriter(file=file) as writer:
                writer.write_table(
                    TableSpec(
                        **{
                            'name': 'foo',
                            'headings': ['a', 'bjørn', 'c'],
                            'rows': [
                                [1, '2', 3],
                                [4, '日本', 6],
                            ]
                        }
                    )
                )

            with zipfile.ZipFile(file.name, 'r') as output_zip:
                with output_zip.open('foo.csv') as csv_file:
                    output = csv.reader(
                        io.TextIOWrapper(csv_file, encoding='utf-8')
                    )

                    assert [row for row in output] == [
                        ['a', 'bjørn', 'c'],
                        ['1', '2', '3'],
                        ['4', '日本', '6'],
                    ]


@pytest.mark.dbtest
class TestSQLWriters(object):

    def _type_convert(self, connection, row):
        """
        Different databases store and return values differently so
        convert the values in the expected row to match the DB.
        """

        def convert(type_map, value):
            func = type_map.get(value.__class__, None)
            return func(value) if func else value

        for driver, type_map in TYPE_MAP.items():
            if driver in connection.engine.driver:
                return {k: convert(type_map, v) for k, v in row.items()}

        return row

    def test_insert(self, writer):
        with writer:
            writer.write_table(
                TableSpec(
                    **{
                        'name': 'foo_insert',
                        'headings': ['id', 'a', 'b', 'c'],
                        'rows': [
                            ['bizzle', 1, 2, 3],
                            ['bazzle', 4, 5, 6],
                        ]
                    }
                )
            )

        # We can use raw SQL instead of SqlAlchemy expressions because
        # we built the DB above
        with writer:
            result = dict([(row['id'], row) for row in writer.connection
                           .execute('SELECT id, a, b, c FROM foo_insert')])

        assert len(result) == 2
        assert dict(result['bizzle']) == {
            'id': 'bizzle',
            'a': 1,
            'b': 2,
            'c': 3
        }
        assert dict(result['bazzle']) == {
            'id': 'bazzle',
            'a': 4,
            'b': 5,
            'c': 6
        }

    def test_upsert(self, writer):
        with writer:
            writer.write_table(
                TableSpec(
                    **{
                        'name': 'foo_upsert',
                        'headings': ['id', 'a', 'b', 'c'],
                        'rows': [['zing', 3, None, 5]]
                    }
                )
            )

        # don't select column 'b' since it hasn't been created yet
        with writer:
            result = dict([
                (row['id'], row) for row in
                writer.connection.execute('SELECT id, a, c FROM foo_upsert')
            ])
        assert len(result) == 1
        assert dict(result['zing']) == {'id': 'zing', 'a': 3, 'c': 5}

        with writer:
            writer.write_table(
                TableSpec(
                    **{
                        'name':
                            'foo_upsert',
                        'headings': ['id', 'a', 'b', 'c'],
                        'rows': [
                            ['bizzle', 1, 'yo', 3],
                            ['bazzle', 4, '日本', 6],
                        ]
                    }
                )
            )

        # We can use raw SQL instead of SqlAlchemy expressions because
        # we built the DB above
        with writer:
            result = dict([(row['id'], row) for row in writer.connection
                           .execute('SELECT id, a, b, c FROM foo_upsert')])

        assert len(result) == 3
        assert dict(result['bizzle']) == {
            'id': 'bizzle',
            'a': 1,
            'b': 'yo',
            'c': 3
        }
        assert dict(result['bazzle']) == {
            'id': 'bazzle',
            'a': 4,
            'b': '日本',
            'c': 6
        }

        with writer:
            writer.write_table(
                TableSpec(
                    **{
                        'name': 'foo_upsert',
                        'headings': ['id', 'a', 'b', 'c'],
                        'rows': [['bizzle', 7, '本', 9],]
                    }
                )
            )

        # We can use raw SQL instead of SqlAlchemy expressions because
        # we built the DB above
        with writer:
            result = dict([(row['id'], row) for row in writer.connection
                           .execute('SELECT id, a, b, c FROM foo_upsert')])

        assert len(result) == 3
        assert dict(result['bizzle']) == {
            'id': 'bizzle',
            'a': 7,
            'b': '本',
            'c': 9
        }
        assert dict(result['bazzle']) == {
            'id': 'bazzle',
            'a': 4,
            'b': '日本',
            'c': 6
        }

    def test_types(self, writer):
        self._test_types(writer, 'foo_fancy_types')

    def _test_types(self, writer, table_name):
        with writer:
            writer.write_table(
                TableSpec(
                    **{
                        'name':
                            table_name,
                        'headings': ['id', 'a', 'b', 'c', 'd', 'e'],
                        'rows': [
                            [
                                'bizzle', 1, 'yo', True,
                                datetime.date(2015, 1, 1),
                                datetime.datetime(2014, 4, 2, 18, 56, 12)
                            ],
                            [
                                'bazzle', 4, '日本', False,
                                datetime.date(2015, 1, 2),
                                datetime.datetime(2014, 5, 1, 11, 16, 45)
                            ],
                        ]
                    }
                )
            )

        # We can use raw SQL instead of SqlAlchemy expressions because
        # we built the DB above
        with writer:
            connection = writer.connection
            result = dict([
                (row['id'], row) for row in connection
                .execute('SELECT id, a, b, c, d, e FROM %s' % table_name)
            ])

            assert len(result) == 2
            expected = {
                'bizzle': {
                    'id': 'bizzle',
                    'a': 1,
                    'b': 'yo',
                    'c': True,
                    'd': datetime.date(2015, 1, 1),
                    'e': datetime.datetime(2014, 4, 2, 18, 56, 12)
                },
                'bazzle': {
                    'id': 'bazzle',
                    'a': 4,
                    'b': '日本',
                    'c': False,
                    'd': datetime.date(2015, 1, 2),
                    'e': datetime.datetime(2014, 5, 1, 11, 16, 45)
                }
            }

            for id, row in result.items():
                assert id in expected
                assert dict(row
                           ) == self._type_convert(connection, expected[id])

    def test_change_type(self, writer):
        self._test_types(writer, 'foo_fancy_type_changes')

        with writer:
            writer.write_table(
                TableSpec(
                    **{
                        'name':
                            'foo_fancy_type_changes',
                        'headings': ['id', 'a', 'b', 'c', 'd', 'e'],
                        'rows': [[
                            'bizzle', 'yo dude', '本', 'true',
                            datetime.datetime(2015, 2, 13),
                            '2014-08-01T11:23:45:00.0000Z'
                        ],]
                    }
                )
            )

        # We can use raw SQL instead of SqlAlchemy expressions because
        # we built the DB above
        with writer:
            result = dict([
                (row['id'], row) for row in writer.connection.execute(
                    'SELECT id, a, b, c, d, e FROM foo_fancy_type_changes'
                )
            ])

        assert len(result) == 2
        expected = {
            'bizzle': {
                'id': 'bizzle',
                'a': 'yo dude',
                'b': '本',
                'c': 'true',
                'd': datetime.date(2015, 2, 13),
                'e': '2014-08-01T11:23:45:00.0000Z'
            },
            'bazzle': {
                'id': 'bazzle',
                'a': '4',
                'b': '日本',
                'c': 'false',
                'd': datetime.date(2015, 1, 2),
                'e': '2014-05-01 11:16:45'
            }
        }

        if 'mysql' in writer.connection.engine.driver:
            # mysql weirdness
            expected['bazzle']['c'] = '0'
        if 'pyodbc' in writer.connection.engine.driver:
            expected['bazzle']['c'] = '0'
            # MSSQL includes fractional seconds in returned value.
            expected['bazzle']['e'] = '2014-05-01 11:16:45.0000000'

        for id, row in result.items():
            assert id in expected
            assert dict(row) == expected[id]

    def test_json_type(self, writer):
        complex_object = {
            'poke1': {
                'name': 'snorlax',
                'color': 'blue',
                'attributes': {
                    'strength': 10,
                    'endurance': 10,
                    'speed': 4,
                },
                'friends': [
                    'pikachu',
                    'charmander',
                ],
            },
            'poke2': {
                'name': 'pikachu',
                'color': 'yellow',
                'attributes': {
                    'strength': 2,
                    'endurance': 2,
                    'speed': 8,
                    'cuteness': 10,
                },
                'friends': [
                    'snorlax',
                    'charmander',
                ],
            },
        }
        with writer:
            if not writer.is_postgres:
                return
            writer.write_table(
                TableSpec(
                    **{
                        'name': 'foo_with_json',
                        'headings': ['id', 'json_col'],
                        'rows': [
                            ['simple', {
                                'k1': 'v1',
                                'k2': 'v2'
                            }],
                            ['with_lists', {
                                'l1': ['i1', 'i2']
                            }],
                            ['complex', complex_object],
                        ],
                        'data_types': [
                            'text',
                            'json',
                        ]
                    }
                )
            )

        # We can use raw SQL instead of SqlAlchemy expressions because
        # we built the DB above
        with writer:
            result = dict([(row['id'], row) for row in writer.connection
                           .execute('SELECT id, json_col FROM foo_with_json')])

        assert len(result) == 3
        assert dict(result['simple']) == {
            'id': 'simple',
            'json_col': {
                'k1': 'v1',
                'k2': 'v2'
            }
        }
        assert dict(result['with_lists']) == {
            'id': 'with_lists',
            'json_col': {
                'l1': ['i1', 'i2']
            }
        }
        assert dict(result['complex']) == {
            'id': 'complex',
            'json_col': complex_object
        }

    def test_explicit_types(self, strict_writer):
        with strict_writer:
            strict_writer.write_table(
                TableSpec(
                    **{
                        'name': 'foo_explicit_types',
                        'headings': ['id', 'a', 'b', 'c', 'd'],
                        'rows': [
                            ['bizzle', '1', 2, 3, '7'],
                            ['bazzle', '4', 5, 6, '8'],
                        ],
                        'data_types': [
                            'text',
                            'integer',
                            'text',
                            None,
                        ]
                    }
                )
            )

        # We can use raw SQL instead of SqlAlchemy expressions because
        # we built the DB above
        with strict_writer:
            result = dict([
                (row['id'], row) for row in strict_writer.connection
                .execute('SELECT id, a, b, c, d FROM foo_explicit_types')
            ])

        assert len(result) == 2
        # a casts strings to ints, b casts ints to text, c default falls back to ints, d default falls back to text
        assert dict(result['bizzle']) == {
            'id': 'bizzle',
            'a': 1,
            'b': '2',
            'c': 3,
            'd': '7'
        }
        assert dict(result['bazzle']) == {
            'id': 'bazzle',
            'a': 4,
            'b': '5',
            'c': 6,
            'd': '8'
        }

    def test_mssql_nvarchar_length_upsize(self, writer):
        with writer:
            if 'odbc' not in writer.connection.engine.driver:
                return

            # Initialize a table with columns where we expect the
            # "some_data" column to be of length 900 bytes, and the
            # "big_data" column to be of nvarchar(max)
            writer.write_table(
                TableSpec(
                    **{
                        'name':
                            'mssql_nvarchar_length',
                        'headings': ['id', 'some_data', 'big_data'],
                        'rows': [
                            [
                                'bizzle', (b'\0' * 800).decode('utf-8'),
                                (b'\0' * 901).decode('utf-8')
                            ],
                            [
                                'bazzle', (b'\0' * 500).decode('utf-8'),
                                (b'\0' * 800).decode('utf-8')
                            ],
                        ]
                    }
                )
            )

            connection = writer.connection

            result = self._get_column_lengths(
                connection, 'mssql_nvarchar_length'
            )
            assert result['some_data'] == ('some_data', 'nvarchar', 900)
            # nvarchar(max) is listed as -1
            assert result['big_data'] == ('big_data', 'nvarchar', -1)

            # put bigger data into "some_column" to ensure it is resized
            # properly
            writer.write_table(
                TableSpec(
                    **{
                        'name':
                            'mssql_nvarchar_length',
                        'headings': ['id', 'some_data', 'big_data'],
                        'rows': [[
                            'sizzle', (b'\0' * 901).decode('utf-8'),
                            (b'\0' * 901).decode('utf-8')
                        ],]
                    }
                )
            )

            result = self._get_column_lengths(
                connection, 'mssql_nvarchar_length'
            )
            assert result['some_data'] == ('some_data', 'nvarchar', -1)
            assert result['big_data'] == ('big_data', 'nvarchar', -1)

    def test_mssql_nvarchar_length_downsize(self, writer):
        with writer:
            if 'odbc' not in writer.connection.engine.driver:
                return

            # Initialize a table with NVARCHAR(max), and make sure
            # smaller data doesn't reduce the size of the column
            metadata = sqlalchemy.MetaData()
            create_sql = sqlalchemy.schema.CreateTable(
                sqlalchemy.Table(
                    'mssql_nvarchar_length_downsize',
                    metadata,
                    sqlalchemy.Column(
                        'id',
                        sqlalchemy.NVARCHAR(length=100),
                        primary_key=True
                    ),
                    sqlalchemy.Column(
                        'some_data', sqlalchemy.NVARCHAR(length=None)
                    ),
                )
            ).compile(writer.connection.engine)
            metadata.create_all(writer.connection.engine)

            writer.write_table(
                TableSpec(
                    **{
                        'name':
                            'mssql_nvarchar_length',
                        'headings': ['id', 'some_data'],
                        'rows': [
                            [
                                'bizzle', (b'\0' * 800).decode('utf-8'),
                                (b'\0' * 800).decode('utf-8')
                            ],
                            [
                                'bazzle', (b'\0' * 500).decode('utf-8'),
                                (b'\0' * 800).decode('utf-8')
                            ],
                        ]
                    }
                )
            )
            result = self._get_column_lengths(
                writer.connection, 'mssql_nvarchar_length_downsize'
            )
            assert result['some_data'] == ('some_data', 'nvarchar', -1)

    def test_big_lump_of_poo(self, writer):
        with writer:
            writer.write_table(
                TableSpec(
                    **{
                        'name': 'foo_with_emoji',
                        'headings': ['id', 'fun_to_be_had'],
                        'rows': [
                            ['A steaming poo', '💩'],
                            ['2020', '😷'],
                        ],
                    }
                )
            )

    def _get_column_lengths(self, connection, table_name):
        return {
            row['COLUMN_NAME']: row for row in connection.execute(
                "SELECT COLUMN_NAME, DATA_TYPE, CHARACTER_MAXIMUM_LENGTH "
                "FROM INFORMATION_SCHEMA.COLUMNS "
                "WHERE TABLE_NAME = '{}';".format(table_name)
            )
        }
